# Copyright (C) 2016 VA Linux Systems Japan K.K.
# Copyright (C) 2016 Fumihiko Kakuma <kakuma at valinux co jp>
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import threading
import time

import netaddr

from os_ken.tests.integrated.common import docker_base as ctn_base
from tempest.common import utils
from tempest import config

from neutron_tempest_plugin.api import base
from neutron_tempest_plugin import config as nconfig
from neutron_tempest_plugin.services.bgp import bgp_client

CONF = config.CONF

Scope = collections.namedtuple('Scope', 'name')
Pool = collections.namedtuple('Pool', 'name, prefixlen, prefixes')
Net = collections.namedtuple('Net', 'name, net, mask, cidr, router')
SubNet = collections.namedtuple('SubNet', 'name, cidr, mask')
Router = collections.namedtuple('Router', 'name, gw')
AS = collections.namedtuple('AS', 'asn, router_id, adv_net')
CHECKTIME = 180
CHECKTIME_INFO = 60
CHECKTIME_INT = 1
BRIDGE_TYPE = ctn_base.BRIDGE_TYPE_DOCKER


def _setup_client_args(auth_provider):
    """Set up ServiceClient arguments using config settings. """
    service = CONF.network.catalog_type or 'network'
    region = CONF.network.region or 'regionOne'
    endpoint_type = CONF.network.endpoint_type
    build_interval = CONF.network.build_interval
    build_timeout = CONF.network.build_timeout

    # The disable_ssl appears in identity
    disable_ssl_certificate_validation = (
        CONF.identity.disable_ssl_certificate_validation)
    ca_certs = CONF.identity.ca_certificates_file

    # Trace in debug section
    trace_requests = CONF.debug.trace_requests

    return [auth_provider, service, region, endpoint_type,
            build_interval, build_timeout,
            disable_ssl_certificate_validation, ca_certs,
            trace_requests]


class BgpSpeakerScenarioTestJSONBase(base.BaseAdminNetworkTest):

    def setUp(self):
        self.addCleanup(self.net_resource_cleanup)
        super(BgpSpeakerScenarioTestJSONBase, self).setUp()

    @classmethod
    def _setup_bgp_non_admin_client(cls):
        mgr = cls.get_client_manager()
        auth_provider = mgr.auth_provider
        client_args = _setup_client_args(auth_provider)
        cls.bgp_client = bgp_client.BgpSpeakerClientJSON(*client_args)

    @classmethod
    def _setup_bgp_admin_client(cls):
        mgr = cls.get_client_manager(credential_type='admin')
        auth_provider = mgr.auth_provider
        client_args = _setup_client_args(auth_provider)
        cls.bgp_adm_client = bgp_client.BgpSpeakerClientJSON(*client_args)

    @classmethod
    def resource_setup(cls):
        super(BgpSpeakerScenarioTestJSONBase, cls).resource_setup()
        if not utils.is_extension_enabled('bgp', 'network'):
            msg = "BGP Speaker extension is not enabled."
            raise cls.skipException(msg)

        cls.images = []
        cls.containers = []
        cls.r_ass = []
        cls.r_as_ip = []
        cls.bridges = []
        cls.admin_routerports = []
        cls.admin_floatingips = []
        cls.admin_routers = []
        cls.admin_router_ip = []
        cls.resource_setup_container()
        cls._setup_bgp_admin_client()
        cls._setup_bgp_non_admin_client()
        cls.lock = threading.Lock()

    @classmethod
    def resource_cleanup(cls):
        for ctn in cls.containers:
            try:
                ctn.stop()
            except ctn_base.CommandError:
                pass
            ctn.remove()
        for br in cls.bridges:
            br.delete()
        super(BgpSpeakerScenarioTestJSONBase, cls).resource_cleanup()

    @classmethod
    def get_subnet(self, start='10.10.1.0', end='10.10.255.0', step=256):
        subnet_gen = netaddr.iter_iprange(start, end, step=step)
        i = 1
        while True:
            with self.lock:
                try:
                    yield (i, str(next(subnet_gen)))
                except StopIteration:
                    subnet_gen = netaddr.iter_iprange(start, end, step=step)
                    yield (i, str(next(subnet_gen)))
                i += 1

    def net_resource_cleanup(self):
        for floatingip in self.admin_floatingips:
            self._try_delete_resource(self.admin_client.delete_floatingip,
                                      floatingip['id'])
        for routerport in self.admin_routerports:
            self._try_delete_resource(
                self.admin_client.remove_router_interface_with_subnet_id,
                routerport['router_id'], routerport['subnet_id'])
        for router in self.admin_routers:
            self._try_delete_resource(self.admin_client.delete_router,
                                      router['id'])

    def create_bgp_speaker(self, auto_delete=True, **args):
        data = {'bgp_speaker': args}
        bgp_speaker = self.bgp_adm_client.create_bgp_speaker(data)
        bgp_speaker_id = bgp_speaker['bgp_speaker']['id']
        if auto_delete:
            self.addCleanup(self.bgp_adm_client.delete_bgp_speaker,
                            bgp_speaker_id)
        if nconfig.CONF.neutron_plugin_options.bgp_schedule_speakers_to_agents:
            self.add_bgp_speaker_to_dragent(self.get_dragent_id(),
                                            bgp_speaker_id)
        return bgp_speaker['bgp_speaker']

    def delete_bgp_speaker(self, id):
        return self.bgp_adm_client.delete_bgp_speaker(id)

    def create_bgp_peer(self, auto_delete=True, **args):
        bgp_peer = self.bgp_adm_client.create_bgp_peer({'bgp_peer': args})
        bgp_peer_id = bgp_peer['bgp_peer']['id']
        if auto_delete:
            self.addCleanup(self.bgp_adm_client.delete_bgp_peer,
                            bgp_peer_id)
        return bgp_peer['bgp_peer']

    def delete_bgp_peer(self, id):
        return self.bgp_adm_client.delete_bgp_peer(id)

    def get_dragent_id(self):
        agents = self.admin_client.list_agents(
            agent_type="BGP dynamic routing agent")
        self.assertTrue(agents['agents'][0]['alive'])
        return agents['agents'][0]['id']

    def add_bgp_speaker_to_dragent(self, agent_id, speaker_id):
        self.bgp_adm_client.add_bgp_speaker_to_dragent(agent_id, speaker_id)

    # tnets[[net1, subnet1, router1], [net2, subnet2, router2], ...]
    def create_bgp_network(self, ip_version, scope,
                           exnet, expool, exsubnet,
                           tpool, tnets):
        addr_scope = self.create_address_scope(scope.name,
                                               ip_version=ip_version)
        # external network
        ext_net = self.create_shared_network(**{'router:external': True})
        ext_net_id = ext_net['id']
        ext_subnetpool = self.create_subnetpool(
            expool.name,
            is_admin=True,
            default_prefixlen=expool.prefixlen,
            address_scope_id=addr_scope['id'],
            prefixes=expool.prefixes)
        self.create_subnet(
            {'id': ext_net_id},
            cidr=netaddr.IPNetwork(exsubnet.cidr),
            mask_bits=exsubnet.mask,
            ip_version=ip_version,
            client=self.admin_client,
            subnetpool_id=ext_subnetpool['id'],
            reserve_cidr=False)
        # tenant network
        tenant_subnetpool = self.create_subnetpool(
            tpool.name,
            default_prefixlen=tpool.prefixlen,
            address_scope_id=addr_scope['id'],
            prefixes=tpool.prefixes)
        for tnet, tsubnet, router in tnets:
            tenant_net = self.create_network()
            tenant_subnet = self.create_subnet(
                {'id': tenant_net['id']},
                cidr=netaddr.IPNetwork(tsubnet.cidr),
                mask_bits=tsubnet.mask,
                ip_version=ip_version,
                subnetpool_id=tenant_subnetpool['id'],
                reserve_cidr=False)
            # router
            ext_gw_info = {'network_id': ext_net_id}
            router_cr = self.admin_client.create_router(
                router.name,
                external_gateway_info=ext_gw_info)['router']
            self.admin_routers.append(router_cr)
            self.admin_client.add_router_interface_with_subnet_id(
                router_cr['id'],
                tenant_subnet['id'])
            self.admin_routerports.append({'router_id': router_cr['id'],
                                           'subnet_id': tenant_subnet['id']})
            router = self.admin_client.show_router(router_cr['id'])['router']
            fixed_ips = router['external_gateway_info']['external_fixed_ips']
            self.admin_router_ip.append(fixed_ips[0]['ip_address'])
        return ext_net_id

    def create_and_add_peers_to_speaker(self, ext_net_id,
                                        speaker_info, peer_infos,
                                        auto_delete=True):
        speaker = self.create_bgp_speaker(auto_delete=auto_delete,
                                          **speaker_info)
        speaker_id = speaker['id']
        self.bgp_adm_client.add_bgp_gateway_network(speaker_id,
                                                    ext_net_id)
        peer_ids = []
        for peer_args in peer_infos:
            peer = self.create_bgp_peer(auto_delete=auto_delete,
                                        **peer_args)
            peer_id = peer['id']
            peer_ids.append(peer_id)
            self.bgp_adm_client.add_bgp_peer_with_id(speaker_id,
                                                     peer_id)
        return (speaker_id, peer_ids)

    def get_remote_as_state(self, l_as, r_as,
                            expected_state,
                            init_state=ctn_base.BGP_FSM_IDLE,
                            checktime=CHECKTIME,
                            checktime_int=CHECKTIME_INT):
        neighbor_state = init_state
        for i in range(0, checktime):
            neighbor_state = r_as.get_neighbor_state(l_as)
            if neighbor_state == expected_state:
                break
            time.sleep(checktime_int)
        return neighbor_state

    def check_remote_as_state(self, l_as, r_as,
                              expected_state,
                              init_state=ctn_base.BGP_FSM_IDLE,
                              checktime=CHECKTIME,
                              checktime_int=CHECKTIME_INT):
        neighbor_state = self.get_remote_as_state(l_as, r_as,
                                                  expected_state,
                                                  init_state=init_state,
                                                  checktime=checktime,
                                                  checktime_int=checktime_int)
        self.assertEqual(neighbor_state, expected_state)

    def get_remote_as_of_state_ok(self, l_as, r_ass,
                                  expected_state,
                                  init_state=ctn_base.BGP_FSM_IDLE,
                                  checktime=CHECKTIME,
                                  checktime_int=CHECKTIME_INT):
        neighbor_state = init_state
        ras_list = []
        ras_max = len(r_ass)
        for r_as in r_ass:
            ras_list.append({'as': r_as, 'check': False})
        ok_ras = []
        for i in range(0, checktime):
            for ras in ras_list:
                if ras['check']:
                    continue
                neighbor_state = ras['as'].get_neighbor_state(l_as)
                if neighbor_state == expected_state:
                    ras['check'] = True
                    ok_ras.append(ras['as'])
            if len(ok_ras) >= ras_max:
                break
            time.sleep(checktime_int)
        return ok_ras

    def check_multi_remote_as_state(self, l_as, r_ass,
                                    expected_state,
                                    init_state=ctn_base.BGP_FSM_IDLE,
                                    checktime=CHECKTIME,
                                    checktime_int=CHECKTIME_INT):
        ok_ras = self.get_remote_as_of_state_ok(
            l_as, r_ass,
            expected_state,
            init_state=init_state,
            checktime=checktime,
            checktime_int=checktime_int)
        self.assertEqual(len(ok_ras), len(r_ass))

    def get_remote_as_rib(self, r_as, prefix, rf, key, expected_item,
                          checktime=CHECKTIME_INFO,
                          checktime_int=CHECKTIME_INT):
        item = None
        for i in range(0, checktime):
            rib = r_as.get_global_rib(prefix=prefix, rf=rf)
            if rib and key in rib[0]:
                if expected_item == rib[0][key]:
                    item = rib[0][key]
                    break
            time.sleep(checktime_int)
        return item

    def check_remote_as_rib(self, r_as, prefix, rf, key, expected_item,
                            checktime=CHECKTIME_INFO,
                            checktime_int=CHECKTIME_INT):
        item = self.get_remote_as_rib(r_as=r_as, prefix=prefix, rf=rf,
                                      key=key, expected_item=expected_item,
                                      checktime=checktime,
                                      checktime_int=checktime_int)
        self.assertEqual(expected_item, item)

    def get_remote_as_of_rib_ok(self, r_ass, prefix, rf, key, expected_item,
                                checktime=CHECKTIME_INFO,
                                checktime_int=CHECKTIME_INT):
        ras_list = []
        ras_max = len(r_ass)
        for r_as in r_ass:
            ras_list.append({'as': r_as, 'check': False})
        ok_ras = []
        for i in range(0, checktime):
            for ras in ras_list:
                if ras['check']:
                    continue
                rib = r_as.get_global_rib(prefix=prefix, rf=rf)
                if rib and key in rib[0]:
                    if expected_item == rib[0][key]:
                        ras['check'] = True
                        ok_ras.append(ras['as'])
            if len(ok_ras) >= ras_max:
                break
            time.sleep(checktime_int)
        return ok_ras

    def check_multi_remote_as_rib(self, r_ass, prefix, rf, key, expected_item,
                                  checktime=CHECKTIME_INFO,
                                  checktime_int=CHECKTIME_INT):
        ok_ras = self.get_remote_as_of_rib_ok(
                   r_ass=r_ass, prefix=prefix, rf=rf,
                   key=key, expected_item=expected_item,
                   checktime=checktime,
                   checktime_int=checktime_int)
        self.assertEqual(len(ok_ras), len(r_ass))

    def get_next_hop(self, speaker_id, dest_addr):
        routes = self.bgp_adm_client.get_bgp_advertised_routes(speaker_id)
        next_hop = ''
        for route in routes['advertised_routes']:
            if route['destination'] == dest_addr:
                next_hop = route['next_hop']
                break
        return next_hop



# Copyright 2012 OpenStack Foundation
# Copyright 2013 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log

from tempest.common import utils
from tempest import config
from tempest.lib.common.utils import data_utils
from tempest.lib.common.utils import test_utils
from tempest.lib import exceptions as lib_exc
from tempest.scenario import manager

CONF = config.CONF

LOG = log.getLogger(__name__)


class ScenarioTest(manager.NetworkScenarioTest):
    """Base class for scenario tests. Uses tempest own clients. """

    credentials = ['primary']


class NetworkScenarioTest(ScenarioTest):
    """Base class for network scenario tests.

    This class provide helpers for network scenario tests, using the neutron
    API. Helpers from ancestor which use the nova network API are overridden
    with the neutron API.

    This Class also enforces using Neutron instead of novanetwork.
    Subclassed tests will be skipped if Neutron is not enabled

    """

    credentials = ['primary', 'admin']

    @classmethod
    def skip_checks(cls):
        super(NetworkScenarioTest, cls).skip_checks()
        if not CONF.service_available.neutron:
            raise cls.skipException('Neutron not available')
        if not utils.is_extension_enabled('bgpvpn', 'network'):
            msg = "Bgpvpn extension not enabled."
            raise cls.skipException(msg)

    def _check_remote_connectivity(self, source, dest, should_succeed=True,
                                   nic=None):
        """check ping server via source ssh connection

        :param source: RemoteClient: an ssh connection from which to ping
        :param dest: and IP to ping against
        :param should_succeed: boolean should ping succeed or not
        :param nic: specific network interface to ping from
        :returns: boolean -- should_succeed == ping
        :returns: ping is false if ping failed
        """
        def ping_remote():
            try:
                source.ping_host(dest, nic=nic)
            except lib_exc.SSHExecCommandFailed:
                LOG.warning('Failed to ping IP: %s via a ssh connection '
                            'from: %s.', dest, source.ssh_client.host)
                return not should_succeed
            return should_succeed

        return test_utils.call_until_true(ping_remote,
                                          CONF.validation.ping_timeout,
                                          1)

    def create_loginable_secgroup_rule(self, security_group_rules_client=None,
                                       secgroup=None,
                                       security_groups_client=None):
        """Create loginable security group rule

        This function will create:
        1. egress and ingress tcp port 22 allow rule in order to allow ssh
        access for ipv4.
        2. egress and ingress tcp port 80 allow rule in order to allow http
        access for ipv4.
        3. egress and ingress ipv6 icmp allow rule, in order to allow icmpv6.
        4. egress and ingress ipv4 icmp allow rule, in order to allow icmpv4.
        """

        if security_group_rules_client is None:
            security_group_rules_client = self.security_group_rules_client
        if security_groups_client is None:
            security_groups_client = self.security_groups_client
        rules = []
        rulesets = [
            dict(
                # ssh
                protocol='tcp',
                port_range_min=22,
                port_range_max=22,
            ),
            dict(
                # http
                protocol='tcp',
                port_range_min=80,
                port_range_max=80,
            ),
            dict(
                # ping
                protocol='icmp',
            ),
            dict(
                # ipv6-icmp for ping6
                protocol='icmp',
                ethertype='IPv6',
            )
        ]
        sec_group_rules_client = security_group_rules_client
        for ruleset in rulesets:
            for r_direction in ['ingress', 'egress']:
                ruleset['direction'] = r_direction
                try:
                    sg_rule = self.create_security_group_rule(
                        sec_group_rules_client=sec_group_rules_client,
                        secgroup=secgroup,
                        security_groups_client=security_groups_client,
                        **ruleset)
                except lib_exc.Conflict as ex:
                    # if rule already exist - skip rule and continue
                    msg = 'Security group rule already exists'
                    if msg not in ex._error_string:
                        raise ex
                else:
                    self.assertEqual(r_direction, sg_rule['direction'])
                    rules.append(sg_rule)

        return rules

    def _create_router(self, client=None, tenant_id=None,
                       namestart='router-smoke'):
        if not client:
            client = self.admin_routers_client
        if not tenant_id:
            tenant_id = client.tenant_id
        name = data_utils.rand_name(namestart)
        result = client.create_router(name=name,
                                      admin_state_up=True,
                                      tenant_id=tenant_id)
        router = result['router']
        self.assertEqual(router['name'], name)
        self.addCleanup(test_utils.call_and_ignore_notfound_exc,
                        client.delete_router,
                        router['id'])
        return router



# Copyright (c) 2018 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import subprocess
import sys

from oslo_log import log
from tempest.lib import exceptions as lib_exc

from neutron_tempest_plugin.common import ssh
from neutron_tempest_plugin import config
from neutron_tempest_plugin import exceptions


LOG = log.getLogger(__name__)

CONF = config.CONF

if ssh.Client.proxy_jump_host:
    # Perform all SSH connections passing through configured SSH server
    SSH_PROXY_CLIENT = ssh.Client.create_proxy_client()
else:
    SSH_PROXY_CLIENT = None


def execute(command, ssh_client=None, timeout=None, check=True):
    """Execute command inside a remote or local shell

    :param command: command string to be executed

    :param ssh_client: SSH client instance used for remote shell execution

    :param timeout: command execution timeout in seconds

    :param check: when False it doesn't raises ShellCommandFailed when
    exit status is not zero. True by default

    :returns: STDOUT text when command execution terminates with zero exit
    status.

    :raises ShellTimeoutExpired: when timeout expires before command execution
    terminates. In such case it kills the process, then it eventually would
    try to read STDOUT and STDERR buffers (not fully implemented) before
    raising the exception.

    :raises ShellCommandFailed: when command execution terminates with non-zero
    exit status.
    """
    ssh_client = ssh_client or SSH_PROXY_CLIENT
    if timeout:
        timeout = float(timeout)

    if ssh_client:
        result = execute_remote_command(command=command, timeout=timeout,
                                        ssh_client=ssh_client)
    else:
        result = execute_local_command(command=command, timeout=timeout)

    if result.exit_status == 0:
        LOG.debug("Command %r succeeded:\n"
                  "stderr:\n%s\n"
                  "stdout:\n%s\n",
                  command, result.stderr, result.stdout)
    elif result.exit_status is None:
        LOG.debug("Command %r timeout expired (timeout=%s):\n"
                  "stderr:\n%s\n"
                  "stdout:\n%s\n",
                  command, timeout, result.stderr, result.stdout)
    else:
        LOG.debug("Command %r failed (exit_status=%s):\n"
                  "stderr:\n%s\n"
                  "stdout:\n%s\n",
                  command, result.exit_status, result.stderr, result.stdout)
    if check:
        result.check()

    return result


def execute_remote_command(command, ssh_client, timeout=None):
    """Execute command on a remote host using SSH client"""
    LOG.debug("Executing command %r on remote host %r (timeout=%r)...",
              command, ssh_client.host, timeout)

    stdout = stderr = exit_status = None

    try:
        # TODO(fressi): re-implement to capture stderr
        stdout = ssh_client.exec_command(command, timeout=timeout)
        exit_status = 0

    except lib_exc.TimeoutException:
        # TODO(fressi): re-implement to capture STDOUT and STDERR and make
        # sure process is killed
        pass

    except lib_exc.SSHExecCommandFailed as ex:
        # Please note class SSHExecCommandFailed has been re-based on
        # top of ShellCommandFailed
        stdout = ex.stdout
        stderr = ex.stderr
        exit_status = ex.exit_status

    return ShellExecuteResult(command=command, timeout=timeout,
                              exit_status=exit_status,
                              stdout=stdout, stderr=stderr)


def execute_local_command(command, timeout=None):
    """Execute command on local host using local shell"""

    LOG.debug("Executing command %r on local host (timeout=%r)...",
              command, timeout)

    process = subprocess.Popen(command, shell=True,
                               universal_newlines=True,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)

    if timeout and sys.version_info < (3, 3):
        # TODO(fressi): re-implement to timeout support on older Pythons
        LOG.warning("Popen.communicate method doens't support for timeout "
                    "on Python %r", sys.version)
        timeout = None

    # Wait for process execution while reading STDERR and STDOUT streams
    if timeout:
        try:
            stdout, stderr = process.communicate(timeout=timeout)
        except subprocess.TimeoutExpired:
            # At this state I expect the process to be still running
            # therefore it has to be kill later after calling poll()
            LOG.exception("Command %r timeout expired.", command)
            stdout = stderr = None
    else:
        stdout, stderr = process.communicate()

    # Check process termination status
    exit_status = process.poll()
    if exit_status is None:
        # The process is still running after calling communicate():
        # let kill it and then read buffers again
        process.kill()
        stdout, stderr = process.communicate()

    return ShellExecuteResult(command=command, timeout=timeout,
                              stdout=stdout, stderr=stderr,
                              exit_status=exit_status)


class ShellExecuteResult(collections.namedtuple(
        'ShellExecuteResult', ['command', 'timeout', 'exit_status', 'stdout',
                               'stderr'])):

    def check(self):
        if self.exit_status is None:
            raise exceptions.ShellTimeoutExpired(command=self.command,
                                                 timeout=self.timeout,
                                                 stderr=self.stderr,
                                                 stdout=self.stdout)

        elif self.exit_status != 0:
            raise exceptions.ShellCommandFailed(command=self.command,
                                                exit_status=self.exit_status,
                                                stderr=self.stderr,
                                                stdout=self.stdout)



# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants as lib_constants
from oslo_log import log
from paramiko import ssh_exception as ssh_exc
from tempest.common import utils as tempest_utils
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc
import testtools

from neutron_tempest_plugin.common import ip
from neutron_tempest_plugin.common import ssh
from neutron_tempest_plugin.common import utils
from neutron_tempest_plugin import config
from neutron_tempest_plugin.scenario import base

CONF = config.CONF

LOG = log.getLogger(__name__)


def turn_nic6_on(ssh, ipv6_port, config_nic=True):
    """Turns the IPv6 vNIC on

    Required because guest images usually set only the first vNIC on boot.
    Searches for the IPv6 vNIC's MAC and brings it up.
    # NOTE(slaweq): on RHEL based OS ifcfg file for new interface is
    # needed to make IPv6 working on it, so if
    # /etc/sysconfig/network-scripts directory exists ifcfg-%(nic)s file
    # should be added in it

    @param ssh: RemoteClient ssh instance to server
    @param ipv6_port: port from IPv6 network attached to the server
    """
    ip_command = ip.IPCommand(ssh)
    nic = ip_command.get_nic_name_by_mac(ipv6_port['mac_address'])

    if config_nic:
        try:
            if sysconfig_network_scripts_dir_exists(ssh):
                ssh.execute_script(
                    'echo -e "DEVICE=%(nic)s\\nNAME=%(nic)s\\nIPV6INIT=yes" | '
                    'tee /etc/sysconfig/network-scripts/ifcfg-%(nic)s; '
                    % {'nic': nic}, become_root=True)
            if nmcli_command_exists(ssh):
                ssh.execute_script('nmcli connection reload %s' % nic,
                                   become_root=True)
                ssh.execute_script('nmcli con mod %s ipv6.addr-gen-mode eui64'
                                   % nic, become_root=True)
                ssh.execute_script('nmcli connection up %s' % nic,
                                   become_root=True)

        except lib_exc.SSHExecCommandFailed as e:
            # NOTE(slaweq): Sometimes it can happen that this SSH command
            # will fail because of some error from network manager in
            # guest os.
            # But even then doing ip link set up below is fine and
            # IP address should be configured properly.
            LOG.debug("Error creating NetworkManager profile. "
                      "Error message: %(error)s",
                      {'error': e})

    ip_command.set_link(nic, "up")


def configure_eth_connection_profile_NM(ssh):
    """Prepare a Network manager profile for ipv6 port

    By default the NetworkManager uses IPv6 privacy
    format it isn't supported by neutron then we create
    a ether profile with eui64 supported format

    @param ssh: RemoteClient ssh instance to server
    """
    # NOTE(ccamposr): on RHEL based OS we need a ether profile with
    # eui64 format
    if nmcli_command_exists(ssh):
        try:
            ssh.execute_script('nmcli connection add type ethernet con-name '
                               'ether ifname "*"', become_root=True)
            ssh.execute_script('nmcli con mod ether ipv6.addr-gen-mode eui64',
                               become_root=True)

        except lib_exc.SSHExecCommandFailed as e:
            # NOTE(slaweq): Sometimes it can happen that this SSH command
            # will fail because of some error from network manager in
            # guest os.
            # But even then doing ip link set up below is fine and
            # IP address should be configured properly.
            LOG.debug("Error creating NetworkManager profile. "
                      "Error message: %(error)s",
                      {'error': e})


def sysconfig_network_scripts_dir_exists(ssh):
    return "False" not in ssh.execute_script(
        'test -d /etc/sysconfig/network-scripts/ || echo "False"')


def nmcli_command_exists(ssh):
    return "False" not in ssh.execute_script(
        'if ! type nmcli > /dev/null ; then echo "False"; fi')


class IPv6Test(base.BaseTempestTestCase):
    credentials = ['primary', 'admin']

    ipv6_ra_mode = 'slaac'
    ipv6_address_mode = 'slaac'

    @classmethod
    def skip_checks(cls):
        super(IPv6Test, cls).skip_checks()
        if not CONF.network_feature_enabled.ipv6:
            raise cls.skipException("IPv6 is not enabled")

    @classmethod
    @tempest_utils.requires_ext(extension="router", service="network")
    def resource_setup(cls):
        super(IPv6Test, cls).resource_setup()
        cls.reserve_external_subnet_cidrs()
        cls._setup_basic_resources()

    @classmethod
    def _setup_basic_resources(cls):
        cls.network = cls.create_network()
        cls.subnet = cls.create_subnet(cls.network)
        cls.router = cls.create_router_by_client()
        cls.create_router_interface(cls.router['id'], cls.subnet['id'])
        cls.keypair = cls.create_keypair()
        cls.secgroup = cls.create_security_group(
            name=data_utils.rand_name('secgroup'))
        cls.create_loginable_secgroup_rule(secgroup_id=cls.secgroup['id'])
        cls.create_pingable_secgroup_rule(secgroup_id=cls.secgroup['id'])

    def _test_ipv6_address_configured(self, ssh_client, vm, ipv6_port):
        ipv6_address = ipv6_port['fixed_ips'][0]['ip_address']
        ip_command = ip.IPCommand(ssh_client)

        def guest_has_address(expected_address):
            ip_addresses = [a.address for a in ip_command.list_addresses()]
            for ip_address in ip_addresses:
                if expected_address in ip_address:
                    return True
            return False
        # Set NIC with IPv6 to be UP and wait until IPv6 address
        # will be configured on this NIC
        turn_nic6_on(ssh_client, ipv6_port, False)
        # And check if IPv6 address will be properly configured
        # on this NIC
        try:
            utils.wait_until_true(
                lambda: guest_has_address(ipv6_address),
                timeout=60)
        except utils.WaitTimeout:
            LOG.debug('Timeout without NM configuration')
        except (lib_exc.SSHTimeout,
                ssh_exc.AuthenticationException) as ssh_e:
            LOG.debug(ssh_e)
            self._log_console_output([vm])
            self._log_local_network_status()
            raise

        if not guest_has_address(ipv6_address):
            try:
                # Set NIC with IPv6 to be UP and wait until IPv6 address
                # will be configured on this NIC
                turn_nic6_on(ssh_client, ipv6_port)
                # And check if IPv6 address will be properly configured
                # on this NIC
                utils.wait_until_true(
                    lambda: guest_has_address(ipv6_address),
                    timeout=90,
                    exception=RuntimeError(
                        "Timed out waiting for IP address {!r} to be "
                        "configured in the VM {!r}.".format(ipv6_address,
                        vm['id'])))
            except (lib_exc.SSHTimeout,
                    ssh_exc.AuthenticationException) as ssh_e:
                LOG.debug(ssh_e)
                self._log_console_output([vm])
                self._log_local_network_status()
                raise

    def _test_ipv6_hotplug(self, ra_mode, address_mode):
        ipv6_networks = [self.create_network() for _ in range(2)]
        for net in ipv6_networks:
            subnet = self.create_subnet(
                network=net, ip_version=6,
                ipv6_ra_mode=ra_mode, ipv6_address_mode=address_mode)
            self.create_router_interface(self.router['id'], subnet['id'])

        server_kwargs = {
            'flavor_ref': CONF.compute.flavor_ref,
            'image_ref': CONF.compute.image_ref,
            'key_name': self.keypair['name'],
            'networks': [
                {'uuid': self.network['id']},
                {'uuid': ipv6_networks[0]['id']}],
            'security_groups': [{'name': self.secgroup['name']}],
        }
        vm = self.create_server(**server_kwargs)['server']
        self.wait_for_server_active(vm)
        self.wait_for_guest_os_ready(vm)
        ipv4_port = self.client.list_ports(
            network_id=self.network['id'],
            device_id=vm['id'])['ports'][0]
        fip = self.create_floatingip(port=ipv4_port)
        ssh_client = ssh.Client(
            fip['floating_ip_address'], CONF.validation.image_ssh_user,
            pkey=self.keypair['private_key'])

        ipv6_port = self.client.list_ports(
            network_id=ipv6_networks[0]['id'],
            device_id=vm['id'])['ports'][0]
        self._test_ipv6_address_configured(ssh_client, vm, ipv6_port)

        # Now remove this port IPv6 port from the VM and attach new one
        self.delete_interface(vm['id'], ipv6_port['id'])

        # And plug VM to the second IPv6 network
        ipv6_port = self.create_port(ipv6_networks[1])
        # Add NetworkManager profile with ipv6 eui64 format to guest OS
        configure_eth_connection_profile_NM(ssh_client)
        self.create_interface(vm['id'], ipv6_port['id'])
        ip.wait_for_interface_status(
            self.os_primary.interfaces_client, vm['id'],
            ipv6_port['id'], lib_constants.PORT_STATUS_ACTIVE,
            ssh_client=ssh_client, mac_address=ipv6_port['mac_address'])
        self._test_ipv6_address_configured(ssh_client, vm, ipv6_port)

    @testtools.skipUnless(CONF.network_feature_enabled.ipv6_subnet_attributes,
                          "DHCPv6 attributes are not enabled.")
    @decorators.idempotent_id('b13e5408-5250-4a42-8e46-6996ce613e91')
    def test_ipv6_hotplug_slaac(self):
        self._test_ipv6_hotplug("slaac", "slaac")

    @testtools.skipUnless(CONF.network_feature_enabled.ipv6_subnet_attributes,
                          "DHCPv6 attributes are not enabled.")
    @decorators.idempotent_id('9aaedbc4-986d-42d5-9177-3e721728e7e0')
    def test_ipv6_hotplug_dhcpv6stateless(self):
        self._test_ipv6_hotplug("dhcpv6-stateless", "dhcpv6-stateless")



# Copyright 2015 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import testtools

import netaddr
from oslo_utils import uuidutils
from tempest.common import utils
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

from neutron_tempest_plugin.api import test_subnetpools


SUBNETPOOL_NAME = 'smoke-subnetpool'


class SubnetPoolsNegativeTestJSON(test_subnetpools.SubnetPoolsTestBase):

    smaller_prefix = u'10.11.12.0/26'

    @decorators.attr(type='negative')
    @decorators.idempotent_id('0212a042-603a-4f46-99e0-e37de9374d30')
    def test_get_non_existent_subnetpool(self):
        non_exist_id = data_utils.rand_name('subnetpool')
        self.assertRaises(lib_exc.NotFound, self.client.show_subnetpool,
                          non_exist_id)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('dc9336e5-f28f-4658-a0b0-cc79e607007d')
    def test_tenant_get_not_shared_admin_subnetpool(self):
        created_subnetpool = self._create_subnetpool(is_admin=True)
        # None-shared admin subnetpool cannot be retrieved by tenant user.
        self.assertRaises(lib_exc.NotFound, self.client.show_subnetpool,
                          created_subnetpool['id'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('5e1f2f86-d81a-498c-82ed-32a49f4dc4d3')
    def test_delete_non_existent_subnetpool(self):
        non_exist_id = data_utils.rand_name('subnetpool')
        self.assertRaises(lib_exc.NotFound, self.client.delete_subnetpool,
                          non_exist_id)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('d1143fe2-212b-4e23-a308-d18f7d8d78d6')
    def test_tenant_create_shared_subnetpool(self):
        # 'shared' subnetpool can only be created by admin.
        self.assertRaises(lib_exc.Forbidden, self._create_subnetpool,
                          is_admin=False, shared=True)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('6ae09d8f-95be-40ed-b1cf-8b850d45bab5')
    @utils.requires_ext(extension='default-subnetpools', service='network')
    def test_tenant_create_default_subnetpool(self):
        # 'default' subnetpool can only be created by admin.
        self.assertRaises(lib_exc.Forbidden, self._create_subnetpool,
                          is_admin=False, is_default=True)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('4be84d30-60ca-4bd3-8512-db5b36ce1378')
    def test_update_non_existent_subnetpool(self):
        non_exist_id = data_utils.rand_name('subnetpool')
        self.assertRaises(lib_exc.NotFound, self.client.update_subnetpool,
                          non_exist_id, name='foo-name')

    @decorators.attr(type='negative')
    @decorators.idempotent_id('e6cd6d87-6173-45dd-bf04-c18ea7ec7537')
    def test_update_subnetpool_not_modifiable_shared(self):
        # 'shared' attributes can be specified during creation.
        # But this attribute is not modifiable after creation.
        created_subnetpool = self._create_subnetpool(is_admin=True)
        pool_id = created_subnetpool['id']
        self.assertRaises(lib_exc.BadRequest, self.client.update_subnetpool,
                          pool_id, shared=True)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('62f7c43b-bff1-4def-8bb7-4754b840aaad')
    def test_update_subnetpool_prefixes_shrink(self):
        # Shrink current subnetpool prefixes is not supported
        created_subnetpool = self._create_subnetpool()
        self.assertRaises(lib_exc.BadRequest,
                          self.client.update_subnetpool,
                          created_subnetpool['id'],
                          prefixes=[self.smaller_prefix])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('fc011824-153e-4469-97ad-9808eb88cae1')
    def test_create_subnet_different_pools_same_network(self):
        network = self.create_network(network_name='smoke-network')
        created_subnetpool = self._create_subnetpool(
            is_admin=True, prefixes=['192.168.0.0/16'])
        subnet = self.create_subnet(
            network, cidr=netaddr.IPNetwork('10.10.10.0/24'), ip_version=4,
            gateway=None, client=self.admin_client)
        # add the subnet created by admin to the cleanUp because
        # the base.py doesn't delete it using the admin client
        self.addCleanup(self.admin_client.delete_subnet, subnet['id'])
        self.assertRaises(lib_exc.BadRequest, self.create_subnet, network,
                          ip_version=4,
                          subnetpool_id=created_subnetpool['id'],
                          client=self.admin_client)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('9589e332-638e-476e-81bd-013d964aa3cb')
    @utils.requires_ext(extension='address-scope', service='network')
    def test_create_subnetpool_associate_invalid_address_scope(self):
        self.assertRaises(lib_exc.BadRequest, self._create_subnetpool,
                          address_scope_id='foo-addr-scope')

    @decorators.attr(type='negative')
    @decorators.idempotent_id('3b6c5942-485d-4964-a560-55608af020b5')
    @utils.requires_ext(extension='address-scope', service='network')
    def test_create_subnetpool_associate_non_exist_address_scope(self):
        self.assertRaises(lib_exc.NotFound, self._create_subnetpool,
                          address_scope_id=uuidutils.generate_uuid())

    @decorators.attr(type='negative')
    @decorators.idempotent_id('2dfb4269-8657-485a-a053-b022e911456e')
    @utils.requires_ext(extension='address-scope', service='network')
    def test_create_subnetpool_associate_address_scope_prefix_intersect(self):
        address_scope = self.create_address_scope(
            name=data_utils.rand_name('smoke-address-scope'),
            ip_version=4)
        addr_scope_id = address_scope['id']
        self._create_subnetpool(address_scope_id=addr_scope_id)
        subnetpool_data = {'name': 'foo-subnetpool',
                           'prefixes': [u'10.11.12.13/24'],
                           'min_prefixlen': '29',
                           'address_scope_id': addr_scope_id}
        self.assertRaises(lib_exc.Conflict, self._create_subnetpool,
                          **subnetpool_data)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('83a19a13-5384-42e2-b579-43fc69c80914')
    @utils.requires_ext(extension='address-scope', service='network')
    def test_create_sp_associate_address_scope_multiple_prefix_intersect(self):
        address_scope = self.create_address_scope(
            name=data_utils.rand_name('smoke-address-scope'),
            ip_version=4)
        addr_scope_id = address_scope['id']
        self._create_subnetpool(prefixes=[u'20.0.0.0/18', u'30.0.0.0/18'],
                                address_scope_id=addr_scope_id)
        prefixes = [u'40.0.0.0/18', u'50.0.0.0/18', u'30.0.0.0/12']
        subnetpool_data = {'name': 'foo-subnetpool',
                           'prefixes': prefixes,
                           'min_prefixlen': '29',
                           'address_scope_id': addr_scope_id}
        self.assertRaises(lib_exc.Conflict, self._create_subnetpool,
                          **subnetpool_data)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('f06d8e7b-908b-4e94-b570-8156be6a4bf1')
    @utils.requires_ext(extension='address-scope', service='network')
    def test_create_subnetpool_associate_address_scope_of_other_owner(self):
        address_scope = self.create_address_scope(
            name=data_utils.rand_name('smoke-address-scope'), is_admin=True,
            ip_version=4)
        self.assertRaises(lib_exc.NotFound, self._create_subnetpool,
                          address_scope_id=address_scope['id'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('3396ec6c-cb80-4ebe-b897-84e904580bdf')
    @testtools.skipIf(
        utils.is_extension_enabled('rbac-address-scope', 'network'),
        reason="Test is outdated starting from Ussuri release."
    )
    @utils.requires_ext(extension='address-scope', service='network')
    def test_tenant_create_subnetpool_associate_shared_address_scope(self):
        address_scope = self.create_address_scope(
            name=data_utils.rand_name('smoke-address-scope'), is_admin=True,
            shared=True, ip_version=4)
        self.assertRaises(lib_exc.BadRequest, self._create_subnetpool,
                          address_scope_id=address_scope['id'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('6d3d9ad5-32d4-4d63-aa00-8c62f73e2881')
    @utils.requires_ext(extension='address-scope', service='network')
    def test_update_subnetpool_associate_address_scope_of_other_owner(self):
        address_scope = self.create_address_scope(
            name=data_utils.rand_name('smoke-address-scope'), is_admin=True,
            ip_version=4)
        address_scope_id = address_scope['id']
        created_subnetpool = self._create_subnetpool(self.client)
        self.assertRaises(lib_exc.NotFound, self.client.update_subnetpool,
                          created_subnetpool['id'],
                          address_scope_id=address_scope_id)

    def _test_update_subnetpool_prefix_intersect_helper(
            self, pool_1_prefixes, pool_2_prefixes, pool_1_updated_prefixes):
        # create two subnet pools associating  to an address scope.
        # Updating the first subnet pool with the prefix intersecting
        # with the second one should be a failure
        address_scope = self.create_address_scope(
            name=data_utils.rand_name('smoke-address-scope'), ip_version=4)
        addr_scope_id = address_scope['id']
        pool_values = {'address_scope_id': addr_scope_id,
                       'prefixes': pool_1_prefixes}
        created_subnetpool_1 = self._create_subnetpool(**pool_values)
        pool_id_1 = created_subnetpool_1['id']
        pool_values = {'address_scope_id': addr_scope_id,
                       'prefixes': pool_2_prefixes}
        self._create_subnetpool(**pool_values)
        # now update the pool_id_1 with the prefix intersecting with
        # pool_id_2
        self.assertRaises(lib_exc.Conflict, self.client.update_subnetpool,
                          pool_id_1, prefixes=pool_1_updated_prefixes)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('96006292-7214-40e0-a471-153fb76e6b31')
    @utils.requires_ext(extension='address-scope', service='network')
    def test_update_subnetpool_prefix_intersect(self):
        pool_1_prefix = [u'20.0.0.0/18']
        pool_2_prefix = [u'20.10.0.0/24']
        pool_1_updated_prefix = [u'20.0.0.0/12']
        self._test_update_subnetpool_prefix_intersect_helper(
            pool_1_prefix, pool_2_prefix, pool_1_updated_prefix)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('4d3f8a79-c530-4e59-9acf-6c05968adbfe')
    @utils.requires_ext(extension='address-scope', service='network')
    def test_update_subnetpool_multiple_prefix_intersect(self):
        pool_1_prefixes = [u'20.0.0.0/18', u'30.0.0.0/18']
        pool_2_prefixes = [u'20.10.0.0/24', u'40.0.0.0/18', '50.0.0.0/18']
        pool_1_updated_prefixes = [u'20.0.0.0/18', u'30.0.0.0/18',
                                   u'50.0.0.0/12']
        self._test_update_subnetpool_prefix_intersect_helper(
            pool_1_prefixes, pool_2_prefixes, pool_1_updated_prefixes)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('7438e49e-1351-45d8-937b-892059fb97f5')
    @utils.requires_ext(extension='address-scope', service='network')
    def test_tenant_update_sp_prefix_associated_with_shared_addr_scope(self):
        address_scope = self.create_address_scope(
            name=data_utils.rand_name('smoke-address-scope'), is_admin=True,
            shared=True, ip_version=4)
        addr_scope_id = address_scope['id']
        pool_values = {'prefixes': [u'20.0.0.0/18', u'30.0.0.0/18']}

        created_subnetpool = self._create_subnetpool(**pool_values)
        pool_id = created_subnetpool['id']
        # associate the subnetpool to the address scope as an admin
        self.admin_client.update_subnetpool(pool_id,
                                            address_scope_id=addr_scope_id)
        body = self.admin_client.show_subnetpool(pool_id)
        self.assertEqual(addr_scope_id,
                         body['subnetpool']['address_scope_id'])

        # updating the subnetpool prefix by the tenant user should fail
        # since the tenant is not the owner of address scope
        update_prefixes = [u'20.0.0.0/18', u'30.0.0.0/18', u'40.0.0.0/18']
        self.assertRaises(lib_exc.BadRequest, self.client.update_subnetpool,
                          pool_id, prefixes=update_prefixes)

        # admin can update the prefixes
        self.admin_client.update_subnetpool(pool_id, prefixes=update_prefixes)
        body = self.admin_client.show_subnetpool(pool_id)
        self.assertEqual(update_prefixes,
                         body['subnetpool']['prefixes'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('648fee7d-a909-4ced-bad3-3a169444c0a8')
    @utils.requires_ext(extension='address-scope', service='network')
    def test_update_subnetpool_associate_address_scope_wrong_ip_version(self):
        address_scope = self.create_address_scope(
            name=data_utils.rand_name('smoke-address-scope'),
            ip_version=6)
        created_subnetpool = self._create_subnetpool()
        self.assertRaises(lib_exc.BadRequest, self.client.update_subnetpool,
                          created_subnetpool['id'],
                          address_scope_id=address_scope['id'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('2f66dc2f-cc32-4caa-91ec-0c0cd7c46d70')
    def test_update_subnetpool_tenant_id(self):
        subnetpool = self._create_subnetpool()
        self.assertRaises(
            lib_exc.BadRequest,
            self.admin_client.update_subnetpool,
            subnetpool['id'],
            tenant_id=self.admin_client.tenant_id,
        )



# Copyright 2017 Fujitsu Limited.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions
import testscenarios

from neutron_tempest_plugin.api import base

load_tests = testscenarios.load_tests_apply_scenarios


class LoggingTestJSON(base.BaseAdminNetworkTest):

    required_extensions = ['logging', 'standard-attr-description']

    @decorators.idempotent_id('8d2e1ba5-455b-4519-a88e-e587002faba6')
    def test_log_lifecycle(self):
        security_group = self.create_security_group()
        name = data_utils.rand_name('test-log')
        description = data_utils.rand_name('test-log-desc')
        log = self.create_log(name=name, description=description,
                              resource_id=security_group['id'],
                              resource_type='security_group', enabled=True)

        # Test 'show log'
        retrieved_log = self.admin_client.show_log(log['id'])['log']
        self.assertEqual(name, retrieved_log['name'])
        self.assertEqual(description, retrieved_log['description'])
        self.assertEqual('security_group', retrieved_log['resource_type'])
        self.assertTrue(retrieved_log['enabled'])

        # Test 'list logs'
        logs = self.admin_client.list_logs()['logs']
        logs_ids = [log_object['id'] for log_object in logs]
        self.assertIn(log['id'], logs_ids)

        # Test 'update log'
        update_description = data_utils.rand_name('test-log')
        self.admin_client.update_log(log['id'],
                                     description=update_description,
                                     enabled=False)
        retrieved_log = self.admin_client.show_log(log['id'])['log']
        self.assertEqual(update_description, retrieved_log['description'])
        self.assertFalse(retrieved_log['enabled'])

        # Test 'delete log'
        self.admin_client.delete_log(log['id'])
        self.assertRaises(exceptions.NotFound,
                          self.admin_client.show_log, log['id'])

    @decorators.idempotent_id('1af6cdab-0eb0-4e13-8027-d89cf1c7a87a')
    def test_list_supported_logging_types(self):
        # List supported logging types
        # Since returned logging types depends on loaded backend drivers
        # this test is checking only if returned keys are same as expected keys
        expected_log_keys = ['type']

        log_types = self.admin_client.list_loggable_resources()
        actual_list_log_types = log_types['loggable_resources']

        # Verify that only required fields present in logging types
        for log_type in actual_list_log_types:
            self.assertEqual(tuple(expected_log_keys), tuple(log_type.keys()))

    @decorators.idempotent_id('1ab4eb2a-76f5-45b9-816b-1aa497a71eea')
    def test_log_deleted_with_corresponding_security_group(self):
        security_group = self.create_security_group()
        name = data_utils.rand_name('test-log')
        log = self.create_log(
            name=name,
            resource_type='security_group',
            resource_id=security_group['id'],
            enabled=True)

        # Ensure log was created
        retrieved_log = self.admin_client.show_log(log['id'])['log']
        self.assertEqual(name, retrieved_log['name'])
        self.assertEqual(security_group['id'], retrieved_log['resource_id'])
        self.assertEqual('security_group', retrieved_log['resource_type'])
        self.assertTrue(retrieved_log['enabled'])

        # Delete SG
        self.delete_security_group(security_group)

        # Ensure log is also deleted
        self.assertRaises(exceptions.NotFound,
                          self.admin_client.show_log, log['id'])



# Copyright 2016 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from tempest.common import utils
from tempest.lib import decorators
import testtools

from neutron_lib import constants
from neutron_tempest_plugin import config
from neutron_tempest_plugin.scenario import base

CONF = config.CONF


class NetworkTestMixin(object):
    def _check_connectivity(self):
        self.check_connectivity(self.fip['floating_ip_address'],
                                CONF.validation.image_ssh_user,
                                self.keypair['private_key'])

    def _check_snat_port_connectivity(self):
        self._check_connectivity()

        # Put the Router_SNAT port down to make sure the traffic flows through
        # Compute node.
        self._put_snat_port_down(self.network['id'])
        self._check_connectivity()

    def _put_snat_port_down(self, network_id):
        port_id = self.client.list_ports(
            network_id=network_id,
            device_owner=constants.DEVICE_OWNER_ROUTER_SNAT)['ports'][0]['id']
        self.os_admin.network_client.update_port(
            port_id, admin_state_up=False)


class NetworkDvrTest(base.BaseTempestTestCase, NetworkTestMixin):
    credentials = ['primary', 'admin']
    force_tenant_isolation = False

    @classmethod
    @utils.requires_ext(extension="dvr", service="network")
    def skip_checks(cls):
        super(NetworkDvrTest, cls).skip_checks()

    @decorators.idempotent_id('3d73ec1a-2ec6-45a9-b0f8-04a283d9d344')
    @testtools.skipUnless(
        CONF.neutron_plugin_options.l3_agent_mode == 'dvr_snat',
        "Need dvr_snat agent mode assumption.")
    def test_vm_reachable_through_compute(self):
        """Check that the VM is reachable through compute node.

        The test is done by putting the SNAT port down on controller node.
        """
        router = self.create_router_by_client(
            distributed=True, tenant_id=self.client.tenant_id, is_admin=True,
            ha=False)
        self.setup_network_and_server(router=router)
        self._check_snat_port_connectivity()



# Copyright 2021 Red Hat, Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators

from neutron_tempest_plugin.common import ssh
from neutron_tempest_plugin.common import utils
from neutron_tempest_plugin import config
from neutron_tempest_plugin import exceptions
from neutron_tempest_plugin.scenario import base


CONF = config.CONF
LOG = log.getLogger(__name__)


# -s0 -l -c5 &> /tmp/tcpdump_out &
def get_receiver_script(result_file, packets_expected):
    """Script that listen icmp echos and write the output on result_file."""
    return """#!/bin/bash
export LC_ALL=en_US.UTF-8
tcpdump -i any -n -v 'icmp[icmptype] = icmp-echoreply or icmp[icmptype] = \
icmp-echo' -s0 -l -c%(packets_expected)d &> %(result_file)s &
    """ % {'result_file': result_file,
           'packets_expected': packets_expected}


def get_sender_script(result_file, receiver_address, completed_message):
    """Script that sends packets to the receiver server."""
    return """#!/bin/bash
export LC_ALL=en_US.UTF-8
ping -c 5 %(address)s
echo '%(completed_message)s' > %(result_file)s &
    """ % {'result_file': result_file,
           'address': receiver_address,
           'completed_message': completed_message}


class MacLearningTest(base.BaseTempestTestCase):

    credentials = ['primary', 'admin']
    force_tenant_isolation = False

    # Import configuration options
    available_type_drivers = (
        CONF.neutron_plugin_options.available_type_drivers)

    completed_message = "Done!"
    output_file = "/tmp/tcpdump_out"
    sender_output_file = "/tmp/sender_out"
    sender_script_file = "/tmp/ping.sh"
    receiver_script_file = "/tmp/traffic.sh"

    @classmethod
    def skip_checks(cls):
        super(MacLearningTest, cls).skip_checks()
        advanced_image_available = (
            CONF.neutron_plugin_options.advanced_image_ref or
            CONF.neutron_plugin_options.default_image_is_advanced)
        if not advanced_image_available:
            skip_reason = "This test requires advanced tools to be executed"
            raise cls.skipException(skip_reason)

    @classmethod
    def resource_setup(cls):
        super(MacLearningTest, cls).resource_setup()

        if CONF.neutron_plugin_options.default_image_is_advanced:
            cls.flavor_ref = CONF.compute.flavor_ref
            cls.image_ref = CONF.compute.image_ref
            cls.username = CONF.validation.image_ssh_user
        else:
            cls.flavor_ref = (
                CONF.neutron_plugin_options.advanced_image_flavor_ref)
            cls.image_ref = CONF.neutron_plugin_options.advanced_image_ref
            cls.username = CONF.neutron_plugin_options.advanced_image_ssh_user

        # Setup basic topology for servers so that we can log into them
        # It's important to keep port security and DHCP disabled for this test
        cls.network = cls.create_network(port_security_enabled=False)
        cls.subnet = cls.create_subnet(cls.network, enable_dhcp=False)
        cls.router = cls.create_router_by_client()
        cls.create_router_interface(cls.router['id'], cls.subnet['id'])

        cls.keypair = cls.create_keypair()

    def _create_server(self):
        name = data_utils.rand_name("maclearning-server")
        server = self.create_server(
            flavor_ref=self.flavor_ref,
            image_ref=self.image_ref,
            key_name=self.keypair['name'], name=name,
            networks=[{'uuid': self.network['id']}],
            config_drive='True')['server']
        self.wait_for_server_active(server)
        self.wait_for_guest_os_ready(server)
        server['port'] = self.client.list_ports(
            network_id=self.network['id'], device_id=server['id'])['ports'][0]
        server['fip'] = self.create_floatingip(port=server['port'])
        server['ssh_client'] = ssh.Client(server['fip']['floating_ip_address'],
                                          self.username,
                                          pkey=self.keypair['private_key'])
        return server

    def _prepare_sender(self, server, address):
        check_script = get_sender_script(self.sender_output_file, address,
                                         self.completed_message)
        self._check_cmd_installed_on_server(server['ssh_client'], server,
                                            'tcpdump')
        server['ssh_client'].execute_script(
            'echo "%s" > %s' % (check_script, self.sender_script_file))

    def _prepare_listener(self, server, n_packets):
        check_script = get_receiver_script(
            result_file=self.output_file,
            packets_expected=n_packets)
        self._check_cmd_installed_on_server(server['ssh_client'], server,
                                            'tcpdump')
        server['ssh_client'].execute_script(
            'echo "%s" > %s' % (check_script, self.receiver_script_file))

    @decorators.idempotent_id('013686ac-23b1-23e4-8361-10b1c98a2861')
    def test_mac_learning_vms_on_same_network(self):
        """Test mac learning works in a network.

        The receiver server will receive all the sent packets.
        The non receiver should not receive any.

        """
        sender = self._create_server()
        receiver = self._create_server()
        non_receiver = self._create_server()

        def check_server_result(server, expected_result, output_file):
            result = server['ssh_client'].execute_script(
                "cat {path} || echo '{path} not exists yet'".format(
                    path=output_file))
            LOG.debug("VM result: %s", result)
            return expected_result in result

        # Prepare the server that is intended to receive the packets
        self._prepare_listener(receiver, 5)

        # Prepare the server that is not intended receive of the packets.
        self._prepare_listener(non_receiver, 2)

        # Run the scripts
        for server in [receiver, non_receiver]:
            server['ssh_client'].execute_script(
                "bash %s" % self.receiver_script_file, become_root=True)

        # Prepare the server that will make the ping.
        target_ip = receiver['port']['fixed_ips'][0]['ip_address']
        self._prepare_sender(sender, address=target_ip)

        LOG.debug("The receiver IP is: %s", target_ip)
        # Run the sender node script
        sender['ssh_client'].execute_script(
                "bash %s" % self.sender_script_file, become_root=True)

        # Check if the message was sent.
        utils.wait_until_true(
            lambda: check_server_result(
                sender, self.completed_message,
                self.sender_output_file),
            exception=RuntimeError(
                "Sender script wasn't executed properly"))

        # Check receiver server
        receiver_expected_result = '5 packets captured'
        utils.wait_until_true(
            lambda: check_server_result(receiver,
                receiver_expected_result, self.output_file),
            exception=RuntimeError(
                'Receiver server did not receive expected packet'))

        # Check the non_receiver server
        non_receiver_expected_result = '0 packets captured'
        try:
            LOG.debug("Try killing non-receiver tcpdump")
            non_receiver['ssh_client'].execute_script(
                "killall tcpdump && sleep 2", become_root=True)
        except exceptions.SSHScriptFailed:
            LOG.debug("Killing tcpdump failed")
            self.assertTrue(check_server_result(non_receiver,
                            non_receiver_expected_result,
                            self.output_file),
                            'Non targeted server received unexpected packets')
            return



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc
import testtools

from neutron_tempest_plugin.api import base
from neutron_tempest_plugin import config

CONF = config.CONF


class FloatingIPAdminTestJSON(base.BaseAdminNetworkTest):
    force_tenant_isolation = True
    credentials = ['primary', 'alt', 'admin']

    @classmethod
    def resource_setup(cls):
        super(FloatingIPAdminTestJSON, cls).resource_setup()
        cls.ext_net_id = CONF.network.public_network_id
        cls.floating_ip = cls.create_floatingip()
        cls.alt_client = cls.os_alt.network_client
        cls.network = cls.create_network()
        cls.subnet = cls.create_subnet(cls.network)
        cls.router = cls.create_router(data_utils.rand_name('router'),
                                       external_network_id=cls.ext_net_id)
        cls.create_router_interface(cls.router['id'], cls.subnet['id'])
        cls.port = cls.create_port(cls.network)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('11116ee9-4e99-5b15-b8e1-aa7df92ca589')
    def test_associate_floating_ip_with_port_from_another_project(self):
        floating_ip = self.create_floatingip()
        project_id = self.create_project()['id']

        port = self.admin_client.create_port(network_id=self.network['id'],
                                             project_id=project_id)
        self.addCleanup(self.admin_client.delete_port, port['port']['id'])
        self.assertRaises(lib_exc.BadRequest,
                          self.client.update_floatingip,
                          floating_ip['id'], port_id=port['port']['id'])

    @testtools.skipUnless(
        CONF.neutron_plugin_options.specify_floating_ip_address_available,
        "Feature for specifying floating IP address is disabled")
    @decorators.idempotent_id('332a8ae4-402e-4b98-bb6f-532e5a87b8e0')
    def test_create_floatingip_with_specified_ip_address(self):
        # other tests may end up stealing the IP before we can use it
        # since it's on the external network so we need to retry if it's
        # in use.
        for _ in range(100):
            fip = self.get_unused_ip(self.ext_net_id, ip_version=4)
            try:
                created_floating_ip = self.create_floatingip(
                    floating_ip_address=fip,
                    client=self.admin_client)
                break
            except lib_exc.Conflict:
                pass
        else:
            self.fail("Could not get an unused IP after 100 attempts")
        self.assertIsNotNone(created_floating_ip['id'])
        self.assertIsNotNone(created_floating_ip['tenant_id'])
        self.assertEqual(created_floating_ip['floating_ip_address'], fip)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time
from urllib import parse as urlparse

from oslo_serialization import jsonutils
from tempest.lib.common import rest_client as service_client
from tempest.lib import exceptions as lib_exc


class NetworkClientJSON(service_client.RestClient):
    """NetworkClientJSON class

    Tempest REST client for Neutron. Uses v2 of the Neutron API, since the
    V1 API has been removed from the code base.

    Implements create, delete, update, list and show for the basic Neutron
    abstractions (networks, sub-networks, routers, ports and floating IP):

    Implements add/remove interface to router using subnet ID / port ID

    It also implements list, show, update and reset for OpenStack Networking
    quotas
    """

    version = '2.0'
    uri_prefix = "v2.0"

    def get_uri(self, plural_name):
        # get service prefix from resource name

        # The following list represents resource names that do not require
        # changing underscore to a hyphen
        hyphen_exceptions = ["service_profiles", "availability_zones"]
        # The following map is used to construct proper URI
        # for the given neutron resource.
        # No need to populate this map if the neutron resource
        # doesn't have a URI prefix.
        service_resource_prefix_map = {
            'metering_labels': 'metering',
            'metering_label_rules': 'metering',
            'policies': 'qos',
            'bandwidth_limit_rules': 'qos',
            'minimum_bandwidth_rules': 'qos',
            'rule_types': 'qos',
            'logs': 'log',
            'loggable_resources': 'log',
        }
        service_prefix = service_resource_prefix_map.get(
            plural_name)
        if plural_name not in hyphen_exceptions:
            plural_name = plural_name.replace("_", "-")
        if service_prefix:
            uri = '%s/%s/%s' % (self.uri_prefix, service_prefix,
                                plural_name)
        else:
            uri = '%s/%s' % (self.uri_prefix, plural_name)
        return uri

    def build_uri(self, plural_name, **kwargs):
        uri = self.get_uri(plural_name)
        if kwargs:
            uri += '?' + urlparse.urlencode(kwargs, doseq=1)
        return uri

    def pluralize(self, resource_name):
        # get plural from map or just add 's'

        # map from resource name to a plural name
        # needed only for those which can't be constructed as name + 's'
        resource_plural_map = {
            'security_groups': 'security_groups',
            'security_group_rules': 'security_group_rules',
            'quotas': 'quotas',
            'qos_policy': 'policies',
            'rbac_policy': 'rbac_policies',
            'network_ip_availability': 'network_ip_availabilities',
        }
        return resource_plural_map.get(resource_name, resource_name + 's')

    def get_uri_with_links(self, plural_name, uri):
        resp, body = self.get(uri)
        result = {plural_name: self.deserialize_list(body)}
        links = self.deserialize_links(body)
        self.expected_success(200, resp.status)
        return links, service_client.ResponseBody(resp, result)

    def _lister(self, plural_name):
        def _list(**filters):
            uri = self.build_uri(plural_name, **filters)
            resp, body = self.get(uri)
            result = {plural_name: self.deserialize_list(body)}
            self.expected_success(200, resp.status)
            return service_client.ResponseBody(resp, result)

        return _list

    def _deleter(self, resource_name):
        def _delete(resource_id):
            plural = self.pluralize(resource_name)
            uri = '%s/%s' % (self.get_uri(plural), resource_id)
            resp, body = self.delete(uri)
            self.expected_success(204, resp.status)
            return service_client.ResponseBody(resp, body)

        return _delete

    def _shower(self, resource_name):
        def _show(resource_id, **fields):
            # fields is a dict which key is 'fields' and value is a
            # list of field's name. An example:
            # {'fields': ['id', 'name']}
            plural = self.pluralize(resource_name)
            if 'details_quotas' in plural:
                details, plural = plural.split('_')
                uri = '%s/%s/%s' % (self.get_uri(plural),
                                    resource_id, details)
            else:
                uri = '%s/%s' % (self.get_uri(plural), resource_id)

            if fields:
                uri += '?' + urlparse.urlencode(fields, doseq=1)
            resp, body = self.get(uri)
            body = self.deserialize_single(body)
            self.expected_success(200, resp.status)
            return service_client.ResponseBody(resp, body)

        return _show

    def _creater(self, resource_name):
        def _create(**kwargs):
            plural = self.pluralize(resource_name)
            uri = self.get_uri(plural)
            post_data = self.serialize({resource_name: kwargs})
            resp, body = self.post(uri, post_data)
            body = self.deserialize_single(body)
            self.expected_success(201, resp.status)
            return service_client.ResponseBody(resp, body)

        return _create

    def _updater(self, resource_name):
        def _update(res_id, **kwargs):
            headers = kwargs.pop('headers', {})
            plural = self.pluralize(resource_name)
            uri = '%s/%s' % (self.get_uri(plural), res_id)
            post_data = self.serialize({resource_name: kwargs})
            resp, body = self.put(uri, post_data, headers=headers)
            body = self.deserialize_single(body)
            self.expected_success(200, resp.status)
            return service_client.ResponseBody(resp, body)

        return _update

    def __getattr__(self, name):
        method_prefixes = ["list_", "delete_", "show_", "create_", "update_"]
        method_functors = [self._lister,
                           self._deleter,
                           self._shower,
                           self._creater,
                           self._updater]
        for index, prefix in enumerate(method_prefixes):
            prefix_len = len(prefix)
            if name[:prefix_len] == prefix:
                return method_functors[index](name[prefix_len:])
        raise AttributeError(name)

    # Subnetpool methods
    def create_subnetpool(self, name, **kwargs):
        subnetpool_data = {'name': name}
        for arg in kwargs:
            subnetpool_data[arg] = kwargs[arg]

        post_data = {'subnetpool': subnetpool_data}
        body = self.serialize_list(post_data, "subnetpools", "subnetpool")
        uri = self.get_uri("subnetpools")
        resp, body = self.post(uri, body)
        body = {'subnetpool': self.deserialize_list(body)}
        self.expected_success(201, resp.status)
        return service_client.ResponseBody(resp, body)

    def get_subnetpool(self, id):
        uri = self.get_uri("subnetpools")
        subnetpool_uri = '%s/%s' % (uri, id)
        resp, body = self.get(subnetpool_uri)
        body = {'subnetpool': self.deserialize_list(body)}
        self.expected_success(200, resp.status)
        return service_client.ResponseBody(resp, body)

    def delete_subnetpool(self, id):
        uri = self.get_uri("subnetpools")
        subnetpool_uri = '%s/%s' % (uri, id)
        resp, body = self.delete(subnetpool_uri)
        self.expected_success(204, resp.status)
        return service_client.ResponseBody(resp, body)

    def list_subnetpools(self, **filters):
        uri = self.get_uri("subnetpools")
        if filters:
            uri = '?'.join([uri, urlparse.urlencode(filters)])
        resp, body = self.get(uri)
        body = {'subnetpools': self.deserialize_list(body)}
        self.expected_success(200, resp.status)
        return service_client.ResponseBody(resp, body)

    def update_subnetpool(self, id, **kwargs):
        subnetpool_data = {}
        for arg in kwargs:
            subnetpool_data[arg] = kwargs[arg]

        post_data = {'subnetpool': subnetpool_data}
        body = self.serialize_list(post_data, "subnetpools", "subnetpool")
        uri = self.get_uri("subnetpools")
        subnetpool_uri = '%s/%s' % (uri, id)
        resp, body = self.put(subnetpool_uri, body)
        body = {'subnetpool': self.deserialize_list(body)}
        self.expected_success(200, resp.status)
        return service_client.ResponseBody(resp, body)

    def add_subnetpool_prefix(self, id, **kwargs):
        return self._subnetpool_prefix_operation(id, 'add_prefixes', kwargs)

    def remove_subnetpool_prefix(self, id, **kwargs):
        return self._subnetpool_prefix_operation(id,
                                                 'remove_prefixes',
                                                 kwargs)

    def _subnetpool_prefix_operation(self, id, operation, op_body):
        uri = self.get_uri("subnetpools")
        op_prefix_uri = '%s/%s/%s' % (uri, id, operation)
        body = jsonutils.dumps(op_body)
        resp, body = self.put(op_prefix_uri, body)
        body = jsonutils.loads(body)
        self.expected_success(200, resp.status)
        return service_client.ResponseBody(resp, body)

    # Common methods that are hard to automate
    def create_bulk_network(self, names, shared=False):
        network_list = [{'name': name, 'shared': shared} for name in names]
        post_data = {'networks': network_list}
        body = self.serialize_list(post_data, "networks", "network")
        uri = self.get_uri("networks")
        resp, body = self.post(uri, body)
        body = {'networks': self.deserialize_list(body)}
        self.expected_success(201, resp.status)
        return service_client.ResponseBody(resp, body)

    def create_bulk_subnet(self, subnet_list):
        post_data = {'subnets': subnet_list}
        body = self.serialize_list(post_data, 'subnets', 'subnet')
        uri = self.get_uri('subnets')
        resp, body = self.post(uri, body)
        body = {'subnets': self.deserialize_list(body)}
        self.expected_success(201, resp.status)
        return service_client.ResponseBody(resp, body)

    def create_bulk_port(self, port_list):
        post_data = {'ports': port_list}
        body = self.serialize_list(post_data, 'ports', 'port')
        uri = self.get_uri('ports')
        resp, body = self.post(uri, body)
        body = {'ports': self.deserialize_list(body)}
        self.expected_success(201, resp.status)
        return service_client.ResponseBody(resp, body)

    def create_bulk_security_groups(self, security_group_list,
                                    stateless=False):
        group_list = [{'security_group': {'name': name}}
                      for name in security_group_list]
        if stateless:
            for group in group_list:
                group['security_group']['stateful'] = False
        post_data = {'security_groups': group_list}
        body = self.serialize_list(post_data, 'security_groups',
                                   'security_group')
        uri = self.get_uri("security-groups")
        resp, body = self.post(uri, body)
        body = {'security_groups': self.deserialize_list(body)}
        self.expected_success(201, resp.status)
        return service_client.ResponseBody(resp, body)

    def wait_for_resource_deletion(self, resource_type, id):
        """Waits for a resource to be deleted."""
        start_time = int(time.time())
        while True:
            if self.is_resource_deleted(resource_type, id):
                return
            if int(time.time()) - start_time >= self.build_timeout:
                raise lib_exc.TimeoutException
            time.sleep(self.build_interval)

    def is_resource_deleted(self, resource_type, id):
        method = 'show_' + resource_type
        try:
            getattr(self, method)(id)
        except AttributeError:
            raise Exception("Unknown resource type %s " % resource_type)
        except lib_exc.NotFound:
            return True
        return False

    def deserialize_single(self, body):
        return jsonutils.loads(body)

    def deserialize_list(self, body):
        res = jsonutils.loads(body)
        # expecting response in form
        # {'resources': [ res1, res2] } => when pagination disabled
        # {'resources': [..], 'resources_links': {}} => if pagination enabled
        for k in res.keys():
            if k.endswith("_links"):
                continue
            return res[k]

    def deserialize_links(self, body):
        res = jsonutils.loads(body)
        # expecting response in form
        # {'resources': [ res1, res2] } => when pagination disabled
        # {'resources': [..], 'resources_links': {}} => if pagination enabled
        for k in res.keys():
            if k.endswith("_links"):
                return {
                    link['rel']: link['href']
                    for link in res[k]
                }
        return {}

    def serialize(self, data):
        return jsonutils.dumps(data)

    def serialize_list(self, data, root=None, item=None):
        return self.serialize(data)

    def update_quotas(self, tenant_id, **kwargs):
        put_body = {'quota': kwargs}
        body = jsonutils.dumps(put_body)
        uri = '%s/quotas/%s' % (self.uri_prefix, tenant_id)
        resp, body = self.put(uri, body)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body['quota'])

    def reset_quotas(self, tenant_id):
        uri = '%s/quotas/%s' % (self.uri_prefix, tenant_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)
        return service_client.ResponseBody(resp, body)

    def create_router(self, name, admin_state_up=True, **kwargs):
        post_body = {'router': kwargs}
        post_body['router']['name'] = name
        post_body['router']['admin_state_up'] = admin_state_up
        body = jsonutils.dumps(post_body)
        uri = '%s/routers' % (self.uri_prefix)
        resp, body = self.post(uri, body)
        self.expected_success(201, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def _update_router(self, router_id, set_enable_snat, **kwargs):
        uri = '%s/routers/%s' % (self.uri_prefix, router_id)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        update_body = {}
        update_body['name'] = kwargs.get('name', body['router']['name'])
        update_body['admin_state_up'] = kwargs.get(
            'admin_state_up', body['router']['admin_state_up'])
        if 'description' in kwargs:
            update_body['description'] = kwargs['description']
        cur_gw_info = body['router']['external_gateway_info']
        if cur_gw_info:
            # TODO(kevinbenton): setting the external gateway info is not
            # allowed for a regular tenant. If the ability to update is also
            # merged, a test case for this will need to be added similar to
            # the SNAT case.
            cur_gw_info.pop('external_fixed_ips', None)
            if not set_enable_snat:
                cur_gw_info.pop('enable_snat', None)
        update_body['external_gateway_info'] = kwargs.get(
            'external_gateway_info', body['router']['external_gateway_info'])
        if 'distributed' in kwargs:
            update_body['distributed'] = kwargs['distributed']
        if 'ha' in kwargs:
            update_body['ha'] = kwargs['ha']
        if 'routes' in kwargs:
            update_body['routes'] = kwargs['routes']
        if 'enable_ndp_proxy' in kwargs:
            update_body['enable_ndp_proxy'] = kwargs['enable_ndp_proxy']
        update_body = dict(router=update_body)
        update_body = jsonutils.dumps(update_body)
        resp, body = self.put(uri, update_body)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def update_router(self, router_id, **kwargs):
        """Update a router leaving enable_snat to its default value."""
        # If external_gateway_info contains enable_snat the request will fail
        # with 404 unless executed with admin client, and therefore we instruct
        # _update_router to not set this attribute
        # NOTE(salv-orlando): The above applies as long as Neutron's default
        # policy is to restrict enable_snat usage to admins only.
        return self._update_router(router_id, set_enable_snat=False, **kwargs)

    def update_router_with_snat_gw_info(self, router_id, **kwargs):
        """Update a router passing also the enable_snat attribute.

        This method must be execute with admin credentials, otherwise the API
        call will return a 404 error.
        """
        return self._update_router(router_id, set_enable_snat=True, **kwargs)

    def add_router_interface_with_subnet_id(self, router_id, subnet_id):
        uri = '%s/routers/%s/add_router_interface' % (self.uri_prefix,
                                                      router_id)
        update_body = {"subnet_id": subnet_id}
        update_body = jsonutils.dumps(update_body)
        resp, body = self.put(uri, update_body)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def add_router_interface_with_port_id(self, router_id, port_id):
        uri = '%s/routers/%s/add_router_interface' % (self.uri_prefix,
                                                      router_id)
        update_body = {"port_id": port_id}
        update_body = jsonutils.dumps(update_body)
        resp, body = self.put(uri, update_body)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def remove_router_interface_with_subnet_id(self, router_id, subnet_id):
        uri = '%s/routers/%s/remove_router_interface' % (self.uri_prefix,
                                                         router_id)
        update_body = {"subnet_id": subnet_id}
        update_body = jsonutils.dumps(update_body)
        resp, body = self.put(uri, update_body)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def remove_router_interface_with_port_id(self, router_id, port_id):
        uri = '%s/routers/%s/remove_router_interface' % (self.uri_prefix,
                                                         router_id)
        update_body = {"port_id": port_id}
        update_body = jsonutils.dumps(update_body)
        resp, body = self.put(uri, update_body)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def list_router_interfaces(self, uuid):
        uri = '%s/ports?device_id=%s' % (self.uri_prefix, uuid)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def remove_router_extra_routes(self, router_id):
        self.update_router(router_id, routes=None)

    def update_agent(self, agent_id, agent_info):
        """Update an agent

        :param agent_info: Agent update information.
        E.g {"admin_state_up": True}
        """
        uri = '%s/agents/%s' % (self.uri_prefix, agent_id)
        agent = {"agent": agent_info}
        body = jsonutils.dumps(agent)
        resp, body = self.put(uri, body)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def delete_agent(self, agent_id):
        uri = '%s/agents/%s' % (self.uri_prefix, agent_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)
        return service_client.ResponseBody(resp, body)

    def list_routers_on_l3_agent(self, agent_id):
        uri = '%s/agents/%s/l3-routers' % (self.uri_prefix, agent_id)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def list_l3_agents_hosting_router(self, router_id):
        uri = '%s/routers/%s/l3-agents' % (self.uri_prefix, router_id)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def add_router_to_l3_agent(self, agent_id, router_id):
        uri = '%s/agents/%s/l3-routers' % (self.uri_prefix, agent_id)
        post_body = {"router_id": router_id}
        body = jsonutils.dumps(post_body)
        resp, body = self.post(uri, body)
        self.expected_success(201, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def remove_router_from_l3_agent(self, agent_id, router_id):
        uri = '%s/agents/%s/l3-routers/%s' % (
            self.uri_prefix, agent_id, router_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)
        return service_client.ResponseBody(resp, body)

    def list_dhcp_agent_hosting_network(self, network_id):
        uri = '%s/networks/%s/dhcp-agents' % (self.uri_prefix, network_id)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def list_networks_hosted_by_one_dhcp_agent(self, agent_id):
        uri = '%s/agents/%s/dhcp-networks' % (self.uri_prefix, agent_id)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def remove_network_from_dhcp_agent(self, agent_id, network_id):
        uri = '%s/agents/%s/dhcp-networks/%s' % (self.uri_prefix, agent_id,
                                                 network_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)
        return service_client.ResponseBody(resp, body)

    def update_extra_routes(self, router_id, nexthop, destination):
        uri = '%s/routers/%s' % (self.uri_prefix, router_id)
        put_body = {
            'router': {
                'routes': [{'nexthop': nexthop,
                            "destination": destination}]
            }
        }
        body = jsonutils.dumps(put_body)
        resp, body = self.put(uri, body)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def delete_extra_routes(self, router_id):
        uri = '%s/routers/%s' % (self.uri_prefix, router_id)
        null_routes = None
        put_body = {
            'router': {
                'routes': null_routes
            }
        }
        body = jsonutils.dumps(put_body)
        resp, body = self.put(uri, body)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def add_extra_routes_atomic(self, router_id, routes):
        uri = '%s/routers/%s/add_extraroutes' % (self.uri_prefix, router_id)
        request_body = {'router': {'routes': routes}}
        resp, response_body = self.put(uri, jsonutils.dumps(request_body))
        self.expected_success(200, resp.status)
        return service_client.ResponseBody(
            resp, jsonutils.loads(response_body))

    def remove_extra_routes_atomic(self, router_id, routes):
        uri = '%s/routers/%s/remove_extraroutes' % (self.uri_prefix, router_id)
        request_body = {'router': {'routes': routes}}
        resp, response_body = self.put(uri, jsonutils.dumps(request_body))
        self.expected_success(200, resp.status)
        return service_client.ResponseBody(
            resp, jsonutils.loads(response_body))

    def add_dhcp_agent_to_network(self, agent_id, network_id):
        post_body = {'network_id': network_id}
        body = jsonutils.dumps(post_body)
        uri = '%s/agents/%s/dhcp-networks' % (self.uri_prefix, agent_id)
        resp, body = self.post(uri, body)
        self.expected_success(201, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def list_qos_policies(self, **filters):
        if filters:
            uri = '%s/qos/policies?%s' % (self.uri_prefix,
                                          urlparse.urlencode(filters))
        else:
            uri = '%s/qos/policies' % self.uri_prefix
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def create_qos_policy(self, name, description=None, shared=False,
                          project_id=None, is_default=False):
        uri = '%s/qos/policies' % self.uri_prefix
        post_data = {
            'policy': {
                'name': name,
                'shared': shared,
                'is_default': is_default
            }
        }
        if description is not None:
            post_data['policy']['description'] = description
        if project_id is not None:
            post_data['policy']['project_id'] = project_id
        resp, body = self.post(uri, self.serialize(post_data))
        body = self.deserialize_single(body)
        self.expected_success(201, resp.status)
        return service_client.ResponseBody(resp, body)

    def update_qos_policy(self, policy_id, **kwargs):
        uri = '%s/qos/policies/%s' % (self.uri_prefix, policy_id)
        post_data = self.serialize({'policy': kwargs})
        resp, body = self.put(uri, post_data)
        body = self.deserialize_single(body)
        self.expected_success(200, resp.status)
        return service_client.ResponseBody(resp, body)

    def create_dscp_marking_rule(self, policy_id, dscp_mark):
        uri = '%s/qos/policies/%s/dscp_marking_rules' % (
            self.uri_prefix, policy_id)
        post_data = self.serialize({
            'dscp_marking_rule': {
                'dscp_mark': dscp_mark
            }
        })
        resp, body = self.post(uri, post_data)
        self.expected_success(201, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def list_dscp_marking_rules(self, policy_id):
        uri = '%s/qos/policies/%s/dscp_marking_rules' % (
            self.uri_prefix, policy_id)
        resp, body = self.get(uri)
        body = self.deserialize_single(body)
        self.expected_success(200, resp.status)
        return service_client.ResponseBody(resp, body)

    def show_dscp_marking_rule(self, policy_id, rule_id):
        uri = '%s/qos/policies/%s/dscp_marking_rules/%s' % (
            self.uri_prefix, policy_id, rule_id)
        resp, body = self.get(uri)
        body = self.deserialize_single(body)
        self.expected_success(200, resp.status)
        return service_client.ResponseBody(resp, body)

    def update_dscp_marking_rule(self, policy_id, rule_id, **kwargs):
        uri = '%s/qos/policies/%s/dscp_marking_rules/%s' % (
            self.uri_prefix, policy_id, rule_id)
        post_data = {'dscp_marking_rule': kwargs}
        resp, body = self.put(uri, jsonutils.dumps(post_data))
        body = self.deserialize_single(body)
        self.expected_success(200, resp.status)
        return service_client.ResponseBody(resp, body)

    def delete_dscp_marking_rule(self, policy_id, rule_id):
        uri = '%s/qos/policies/%s/dscp_marking_rules/%s' % (
            self.uri_prefix, policy_id, rule_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)
        return service_client.ResponseBody(resp, body)

    def list_qos_rule_types(self):
        uri = '%s/qos/rule-types' % self.uri_prefix
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def show_qos_rule_type(self, rule_type_name):
        uri = '%s/qos/rule-types/%s' % (
            self.uri_prefix, rule_type_name)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def create_trunk(self, parent_port_id=None, subports=None,
                     tenant_id=None, name=None, admin_state_up=None,
                     description=None, **kwargs):
        uri = '%s/trunks' % self.uri_prefix
        if parent_port_id:
            kwargs['port_id'] = parent_port_id
        if subports is not None:
            kwargs['sub_ports'] = subports
        if tenant_id is not None:
            kwargs['tenant_id'] = tenant_id
        if name is not None:
            kwargs['name'] = name
        if description is not None:
            kwargs['description'] = description
        if admin_state_up is not None:
            kwargs['admin_state_up'] = admin_state_up
        resp, body = self.post(uri, self.serialize({'trunk': kwargs}))
        body = self.deserialize_single(body)
        self.expected_success(201, resp.status)
        return service_client.ResponseBody(resp, body)

    def update_trunk(self, trunk_id, **kwargs):
        put_body = {'trunk': kwargs}
        body = jsonutils.dumps(put_body)
        uri = '%s/trunks/%s' % (self.uri_prefix, trunk_id)
        resp, body = self.put(uri, body)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def show_trunk(self, trunk_id):
        uri = '%s/trunks/%s' % (self.uri_prefix, trunk_id)
        resp, body = self.get(uri)
        body = self.deserialize_single(body)
        self.expected_success(200, resp.status)
        return service_client.ResponseBody(resp, body)

    def list_trunks(self, **kwargs):
        uri = '%s/trunks' % self.uri_prefix
        if kwargs:
            uri += '?' + urlparse.urlencode(kwargs, doseq=1)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = self.deserialize_single(body)
        return service_client.ResponseBody(resp, body)

    def delete_trunk(self, trunk_id):
        uri = '%s/trunks/%s' % (self.uri_prefix, trunk_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)
        return service_client.ResponseBody(resp, body)

    def _subports_action(self, action, trunk_id, subports):
        uri = '%s/trunks/%s/%s' % (self.uri_prefix, trunk_id, action)
        resp, body = self.put(uri, jsonutils.dumps({'sub_ports': subports}))
        body = self.deserialize_single(body)
        self.expected_success(200, resp.status)
        return service_client.ResponseBody(resp, body)

    def add_subports(self, trunk_id, subports):
        return self._subports_action('add_subports', trunk_id, subports)

    def remove_subports(self, trunk_id, subports):
        return self._subports_action('remove_subports', trunk_id, subports)

    def get_subports(self, trunk_id):
        uri = '%s/trunks/%s/%s' % (self.uri_prefix, trunk_id, 'get_subports')
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def get_auto_allocated_topology(self, tenant_id=None):
        uri = '%s/auto-allocated-topology/%s' % (self.uri_prefix, tenant_id)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def delete_auto_allocated_topology(self, tenant_id=None):
        uri = '%s/auto-allocated-topology/%s' % (self.uri_prefix, tenant_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)
        return service_client.ResponseBody(resp, body)

    def create_flavor_service_profile(self, flavor_id, service_profile_id):
        body = jsonutils.dumps({'service_profile': {'id': service_profile_id}})
        uri = '%s/flavors/%s/service_profiles' % (self.uri_prefix, flavor_id)
        resp, body = self.post(uri, body)
        self.expected_success(201, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def list_flavor_service_profiles(self, flavor_id):
        uri = '%s/flavors/%s/service_profiles' % (self.uri_prefix, flavor_id)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def delete_flavor_service_profile(self, flavor_id, service_profile_id):
        uri = '%s/flavors/%s/service_profiles/%s' % (self.uri_prefix,
                                                     flavor_id,
                                                     service_profile_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)
        return service_client.ResponseBody(resp, body)

    def list_security_group_rules(self, **kwargs):
        uri = '%s/security-group-rules' % self.uri_prefix
        if kwargs:
            uri += '?' + urlparse.urlencode(kwargs, doseq=1)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def create_security_group_rule(self, direction, security_group_id,
                                   **kwargs):
        post_body = {'security_group_rule': kwargs}
        post_body['security_group_rule']['direction'] = direction
        post_body['security_group_rule'][
            'security_group_id'] = security_group_id
        body = jsonutils.dumps(post_body)
        uri = '%s/security-group-rules' % self.uri_prefix
        resp, body = self.post(uri, body)
        self.expected_success(201, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def delete_security_group_rule(self, security_group_rule_id):
        uri = '%s/security-group-rules/%s' % (self.uri_prefix,
                                              security_group_rule_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)
        return service_client.ResponseBody(resp, body)

    def list_security_groups(self, **kwargs):
        uri = '%s/security-groups' % self.uri_prefix
        if kwargs:
            uri += '?' + urlparse.urlencode(kwargs, doseq=1)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def delete_security_group(self, security_group_id):
        uri = '%s/security-groups/%s' % (
            self.uri_prefix, security_group_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)
        return service_client.ResponseBody(resp, body)

    def list_ports(self, **kwargs):
        uri = '%s/ports' % self.uri_prefix
        if kwargs:
            uri += '?' + urlparse.urlencode(kwargs, doseq=1)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def list_floatingips(self, **kwargs):
        uri = '%s/floatingips' % self.uri_prefix
        if kwargs:
            uri += '?' + urlparse.urlencode(kwargs, doseq=1)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def create_floatingip(self, floating_network_id, **kwargs):
        post_body = {'floatingip': {
            'floating_network_id': floating_network_id}}
        if kwargs:
            post_body['floatingip'].update(kwargs)
        body = jsonutils.dumps(post_body)
        uri = '%s/floatingips' % self.uri_prefix
        resp, body = self.post(uri, body)
        self.expected_success(201, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def get_floatingip(self, fip_id):
        uri = '%s/floatingips/%s' % (self.uri_prefix, fip_id)
        get_resp, get_resp_body = self.get(uri)
        self.expected_success(200, get_resp.status)
        body = jsonutils.loads(get_resp_body)
        return service_client.ResponseBody(get_resp, body)

    def update_floatingip(self, fip_id, **kwargs):
        uri = '%s/floatingips/%s' % (self.uri_prefix, fip_id)
        get_resp, _ = self.get(uri)
        self.expected_success(200, get_resp.status)
        put_body = jsonutils.dumps({'floatingip': kwargs})
        put_resp, resp_body = self.put(uri, put_body)
        self.expected_success(200, put_resp.status)
        body = jsonutils.loads(resp_body)
        return service_client.ResponseBody(put_resp, body)

    def create_port_forwarding(self, fip_id, internal_port_id,
                               internal_port, external_port,
                               internal_ip_address=None, protocol='tcp'):
        post_body = {'port_forwarding': {
            'protocol': protocol,
            'internal_port_id': internal_port_id,
            'internal_port': int(internal_port),
            'external_port': int(external_port)}}
        if internal_ip_address:
            post_body['port_forwarding']['internal_ip_address'] = (
                internal_ip_address)
        body = jsonutils.dumps(post_body)
        uri = '%s/floatingips/%s/port_forwardings' % (self.uri_prefix, fip_id)
        resp, body = self.post(uri, body)
        self.expected_success(201, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def get_port_forwarding(self, fip_id, pf_id):
        uri = '%s/floatingips/%s/port_forwardings/%s' % (self.uri_prefix,
                                                         fip_id, pf_id)
        get_resp, get_resp_body = self.get(uri)
        self.expected_success(200, get_resp.status)
        body = jsonutils.loads(get_resp_body)
        return service_client.ResponseBody(get_resp, body)

    def list_port_forwardings(self, fip_id):
        uri = '%s/floatingips/%s/port_forwardings' % (self.uri_prefix, fip_id)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def update_port_forwarding(self, fip_id, pf_id, **kwargs):
        uri = '%s/floatingips/%s/port_forwardings/%s' % (self.uri_prefix,
                                                         fip_id, pf_id)
        put_body = jsonutils.dumps({'port_forwarding': kwargs})
        put_resp, resp_body = self.put(uri, put_body)
        self.expected_success(200, put_resp.status)
        body = jsonutils.loads(resp_body)
        return service_client.ResponseBody(put_resp, body)

    def delete_port_forwarding(self, fip_id, pf_id):
        uri = '%s/floatingips/%s/port_forwardings/%s' % (self.uri_prefix,
                                                         fip_id, pf_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)
        service_client.ResponseBody(resp, body)

    def create_local_ip(self, network_id, **kwargs):
        post_body = {'local_ip': {
            'network_id': network_id}}
        if kwargs:
            post_body['local_ip'].update(kwargs)
        body = jsonutils.dumps(post_body)
        uri = '%s/local_ips' % self.uri_prefix
        resp, body = self.post(uri, body)
        self.expected_success(201, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def list_local_ips(self, **kwargs):
        uri = '%s/local_ips' % self.uri_prefix
        if kwargs:
            uri += '?' + urlparse.urlencode(kwargs, doseq=1)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def get_local_ip(self, local_ip_id):
        uri = '%s/local_ips/%s' % (self.uri_prefix, local_ip_id)
        get_resp, get_resp_body = self.get(uri)
        self.expected_success(200, get_resp.status)
        body = jsonutils.loads(get_resp_body)
        return service_client.ResponseBody(get_resp, body)

    def update_local_ip(self, local_ip_id, **kwargs):
        uri = '%s/local_ips/%s' % (self.uri_prefix, local_ip_id)
        get_resp, _ = self.get(uri)
        self.expected_success(200, get_resp.status)
        put_body = jsonutils.dumps({'local_ip': kwargs})
        put_resp, resp_body = self.put(uri, put_body)
        self.expected_success(200, put_resp.status)
        body = jsonutils.loads(resp_body)
        return service_client.ResponseBody(put_resp, body)

    def delete_local_ip(self, local_ip_id):
        uri = '%s/local_ips/%s' % (
            self.uri_prefix, local_ip_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)
        return service_client.ResponseBody(resp, body)

    def create_local_ip_association(self, local_ip_id, fixed_port_id,
                                    fixed_ip=None):
        post_body = {'port_association': {
            'fixed_port_id': fixed_port_id}}
        if fixed_ip:
            post_body['port_association']['fixed_ip'] = (
                fixed_ip)
        body = jsonutils.dumps(post_body)
        uri = '%s/local_ips/%s/port_associations' % (self.uri_prefix,
                                                     local_ip_id)
        resp, body = self.post(uri, body)
        self.expected_success(201, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def get_local_ip_association(self, local_ip_id, fixed_port_id):
        uri = '%s/local_ips/%s/port_associations/%s' % (self.uri_prefix,
                                                        local_ip_id,
                                                        fixed_port_id)
        get_resp, get_resp_body = self.get(uri)
        self.expected_success(200, get_resp.status)
        body = jsonutils.loads(get_resp_body)
        return service_client.ResponseBody(get_resp, body)

    def list_local_ip_associations(self, local_ip_id):
        uri = '%s/local_ips/%s/port_associations' % (self.uri_prefix,
                                                     local_ip_id)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def delete_local_ip_association(self, local_ip_id, fixed_port_id):

        uri = '%s/local_ips/%s/port_associations/%s' % (self.uri_prefix,
                                                        local_ip_id,
                                                        fixed_port_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)
        service_client.ResponseBody(resp, body)

    def create_conntrack_helper(self, router_id, helper, protocol, port):
        post_body = {'conntrack_helper': {
            'helper': helper,
            'protocol': protocol,
            'port': port}}
        body = jsonutils.dumps(post_body)
        uri = '%s/routers/%s/conntrack_helpers' % (self.uri_prefix, router_id)
        resp, body = self.post(uri, body)
        self.expected_success(201, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def get_conntrack_helper(self, router_id, cth_id):
        uri = '%s/routers/%s/conntrack_helpers/%s' % (self.uri_prefix,
                                                      router_id, cth_id)
        get_resp, get_resp_body = self.get(uri)
        self.expected_success(200, get_resp.status)
        body = jsonutils.loads(get_resp_body)
        return service_client.ResponseBody(get_resp, body)

    def list_conntrack_helpers(self, router_id):
        uri = '%s/routers/%s/conntrack_helpers' % (self.uri_prefix, router_id)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def update_conntrack_helper(self, router_id, cth_id, **kwargs):
        uri = '%s/routers/%s/conntrack_helpers/%s' % (self.uri_prefix,
                                                      router_id, cth_id)
        put_body = jsonutils.dumps({'conntrack_helper': kwargs})
        put_resp, resp_body = self.put(uri, put_body)
        self.expected_success(200, put_resp.status)
        body = jsonutils.loads(resp_body)
        return service_client.ResponseBody(put_resp, body)

    def delete_conntrack_helper(self, router_id, cth_id):
        uri = '%s/routers/%s/conntrack_helpers/%s' % (self.uri_prefix,
                                                      router_id, cth_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)
        service_client.ResponseBody(resp, body)

    def list_extensions(self, **filters):
        uri = self.get_uri("extensions")
        if filters:
            uri = '?'.join([uri, urlparse.urlencode(filters)])
        resp, body = self.get(uri)
        body = {'extensions': self.deserialize_list(body)}
        self.expected_success(200, resp.status)
        return service_client.ResponseBody(resp, body)

    def get_tags(self, resource_type, resource_id):
        uri = '%s/%s/%s/tags' % (
            self.uri_prefix, resource_type, resource_id)
        resp, body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def get_tag(self, resource_type, resource_id, tag):
        uri = '%s/%s/%s/tags/%s' % (
            self.uri_prefix, resource_type, resource_id, tag)
        resp, body = self.get(uri)
        self.expected_success(204, resp.status)

    def update_tag(self, resource_type, resource_id, tag):
        uri = '%s/%s/%s/tags/%s' % (
            self.uri_prefix, resource_type, resource_id, tag)
        resp, body = self.put(uri, None)
        self.expected_success(201, resp.status)

    def update_tags(self, resource_type, resource_id, tags):
        uri = '%s/%s/%s/tags' % (
            self.uri_prefix, resource_type, resource_id)
        req_body = jsonutils.dumps({'tags': tags})
        resp, body = self.put(uri, req_body)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(body)
        return service_client.ResponseBody(resp, body)

    def delete_tags(self, resource_type, resource_id):
        uri = '%s/%s/%s/tags' % (
            self.uri_prefix, resource_type, resource_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)

    def delete_tag(self, resource_type, resource_id, tag):
        uri = '%s/%s/%s/tags/%s' % (
            self.uri_prefix, resource_type, resource_id, tag)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)

    def add_addresses_to_address_group(self, address_group_id, addresses):
        uri = '%s/address-groups/%s/add_addresses' % (
            self.uri_prefix, address_group_id)
        request_body = {'addresses': addresses}
        resp, response_body = self.put(uri, jsonutils.dumps(request_body))
        self.expected_success(200, resp.status)
        return service_client.ResponseBody(
            resp, jsonutils.loads(response_body))

    def remove_addresses_from_address_group(self, address_group_id, addresses):
        uri = '%s/address-groups/%s/remove_addresses' % (
            self.uri_prefix, address_group_id)
        request_body = {'addresses': addresses}
        resp, response_body = self.put(uri, jsonutils.dumps(request_body))
        self.expected_success(200, resp.status)
        return service_client.ResponseBody(
            resp, jsonutils.loads(response_body))

    def create_ndp_proxy(self, **kwargs):
        uri = '%s/ndp_proxies' % self.uri_prefix
        post_body = jsonutils.dumps({'ndp_proxy': kwargs})
        resp, response_body = self.post(uri, post_body)
        self.expected_success(201, resp.status)
        body = jsonutils.loads(response_body)
        return service_client.ResponseBody(resp, body)

    def list_ndp_proxies(self, **kwargs):
        uri = '%s/ndp_proxies' % self.uri_prefix
        if kwargs:
            uri += '?' + urlparse.urlencode(kwargs, doseq=1)
        resp, response_body = self.get(uri)
        self.expected_success(200, resp.status)
        body = jsonutils.loads(response_body)
        return service_client.ResponseBody(resp, body)

    def get_ndp_proxy(self, ndp_proxy_id):
        uri = '%s/ndp_proxies/%s' % (self.uri_prefix, ndp_proxy_id)
        get_resp, response_body = self.get(uri)
        self.expected_success(200, get_resp.status)
        body = jsonutils.loads(response_body)
        return service_client.ResponseBody(get_resp, body)

    def update_ndp_proxy(self, ndp_proxy_id, **kwargs):
        uri = '%s/ndp_proxies/%s' % (self.uri_prefix, ndp_proxy_id)
        get_resp, _ = self.get(uri)
        self.expected_success(200, get_resp.status)
        put_body = jsonutils.dumps({'ndp_proxy': kwargs})
        put_resp, response_body = self.put(uri, put_body)
        self.expected_success(200, put_resp.status)
        body = jsonutils.loads(response_body)
        return service_client.ResponseBody(put_resp, body)

    def delete_ndp_proxy(self, ndp_proxy_id):
        uri = '%s/ndp_proxies/%s' % (
            self.uri_prefix, ndp_proxy_id)
        resp, body = self.delete(uri)
        self.expected_success(204, resp.status)
        return service_client.ResponseBody(resp, body)



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
from tempest.common import utils
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

from neutron_tempest_plugin.api import base

FAKE_IP = '10.0.0.1'
FAKE_MAC = '00:25:64:e8:19:dd'


@ddt.ddt
class PortSecTest(base.BaseNetworkTest):

    @decorators.idempotent_id('7c338ddf-e64e-4118-bd33-e49a1f2f1495')
    @utils.requires_ext(extension='port-security', service='network')
    def test_port_sec_default_value(self):
        # Default port-sec value is True, and the attr of the port will inherit
        # from the port-sec of the network when it not be specified in API
        network = self.create_network()
        self.assertTrue(network['port_security_enabled'])
        self.create_subnet(network)
        port = self.create_port(network)
        self.assertTrue(port['port_security_enabled'])

    @decorators.idempotent_id('e60eafd2-31de-4c38-8106-55447d033b57')
    @utils.requires_ext(extension='port-security', service='network')
    @ddt.unpack
    @ddt.data({'port_sec_net': False, 'port_sec_port': True, 'expected': True},
              {'port_sec_net': True, 'port_sec_port': False,
               'expected': False})
    def test_port_sec_specific_value(self, port_sec_net, port_sec_port,
                                     expected):
        network = self.create_network(port_security_enabled=port_sec_net)
        self.create_subnet(network)
        port = self.create_port(network, port_security_enabled=port_sec_port)
        self.assertEqual(network['port_security_enabled'], port_sec_net)
        self.assertEqual(port['port_security_enabled'], expected)

    @decorators.idempotent_id('fe7c27b9-f320-4daf-b977-b1547c43daf6')
    @utils.requires_ext(extension='port-security', service='network')
    def test_create_port_sec_with_security_group(self):
        network = self.create_network(port_security_enabled=True)
        self.create_subnet(network)

        port = self.create_port(network, security_groups=[])
        self.assertTrue(port['port_security_enabled'])
        self.client.delete_port(port['id'])

        port = self.create_port(network, security_groups=[],
                                port_security_enabled=False)
        self.assertFalse(port['port_security_enabled'])
        self.assertEmpty(port['security_groups'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('ff11226c-a5ff-4ad4-8480-0840e36e47a9')
    @utils.requires_ext(extension='port-security', service='network')
    def test_port_sec_update_port_failed(self):
        network = self.create_network()
        self.create_subnet(network)

        security_group = self.create_security_group()
        port = self.create_port(network)

        # Exception when set port-sec to False with sec-group defined
        self.assertRaises(lib_exc.Conflict, self.update_port, port,
                          port_security_enabled=False)

        port = self.update_port(port, security_groups=[],
                                port_security_enabled=False)
        self.assertEmpty(port['security_groups'])
        self.assertFalse(port['port_security_enabled'])
        port = self.update_port(
            port, security_groups=[security_group['id']],
            port_security_enabled=True)

        self.assertNotEmpty(port['security_groups'])
        self.assertTrue(port['port_security_enabled'])

        # Remove security group from port before deletion on resource_cleanup
        self.update_port(port, security_groups=[])

    @decorators.idempotent_id('05642059-1bfc-4581-9bc9-aaa5db08dd60')
    @utils.requires_ext(extension='port-security', service='network')
    def test_port_sec_update_pass(self):
        network = self.create_network()
        self.create_subnet(network)
        security_group = self.create_security_group()

        port = self.create_port(network,
                                security_groups=[security_group['id']],
                                port_security_enabled=True)
        self.assertNotEmpty(port['security_groups'])
        self.assertTrue(port['port_security_enabled'])

        port = self.update_port(port, security_groups=[])
        self.assertEmpty(port['security_groups'])
        self.assertTrue(port['port_security_enabled'])

        port = self.update_port(port, security_groups=[security_group['id']])
        self.assertNotEmpty(port['security_groups'])
        port = self.update_port(port, security_groups=[],
                                port_security_enabled=False)
        self.assertEmpty(port['security_groups'])
        self.assertFalse(port['port_security_enabled'])

    @decorators.idempotent_id('2df6114b-b8c3-48a1-96e8-47f08159d35c')
    @utils.requires_ext(extension='port-security', service='network')
    def test_delete_with_port_sec(self):
        network = self.create_network(port_security_enabled=True)
        port = self.create_port(network=network,
                                port_security_enabled=True)
        self.client.delete_port(port['id'])
        self.assertTrue(self.client.is_resource_deleted('port', port['id']))
        self.client.delete_network(network['id'])
        self.assertTrue(
            self.client.is_resource_deleted('network', network['id']))

    @decorators.attr(type='negative')
    @decorators.idempotent_id('ed93e453-3f8d-495e-8e7e-b0e268c2ebd9')
    @utils.requires_ext(extension='port-security', service='network')
    @utils.requires_ext(extension='allowed-address-pairs', service='network')
    def test_allowed_address_pairs(self):
        network = self.create_network()
        self.create_subnet(network)
        port = self.create_port(network=network, port_security_enabled=False)
        allowed_address_pairs = [{'ip_address': FAKE_IP,
                                  'mac_address': FAKE_MAC}]

        # Exception when set address-pairs with port-sec is False
        self.assertRaises(lib_exc.Conflict,
                          self.update_port, port,
                          allowed_address_pairs=allowed_address_pairs)



# Copyright 2020 Ericsson Software Technology
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections

from neutron_lib import constants as nlib_const
from oslo_log import log as logging
from tempest.common import utils
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators
from tempest.lib import exceptions
import testtools

from neutron_tempest_plugin.common import ssh
from neutron_tempest_plugin import config
from neutron_tempest_plugin.scenario import base

LOG = logging.getLogger(__name__)
CONF = config.CONF

Server = collections.namedtuple(
    'Server', ['floating_ip', 'server', 'ssh_client'])


class MetadataTest(base.BaseTempestTestCase):

    """Test metadata access over IPv6 tenant subnet.

    Please note that there is metadata over IPv4 test coverage in tempest:

    tempest.scenario.test_server_basic_ops\
        .TestServerBasicOps.test_server_basic_ops
    """

    credentials = ['primary', 'admin']
    force_tenant_isolation = False

    @classmethod
    def skip_checks(cls):
        super(MetadataTest, cls).skip_checks()
        if not utils.is_network_feature_enabled('ipv6_metadata'):
            raise cls.skipException("Metadata over IPv6 is not enabled")

    @classmethod
    def resource_setup(cls):
        super(MetadataTest, cls).resource_setup()
        cls.rand_name = data_utils.rand_name(
            cls.__name__.rsplit('.', 1)[-1])
        cls.reserve_external_subnet_cidrs()
        cls.network = cls.create_network(name=cls.rand_name)
        cls.subnet_v4 = cls.create_subnet(
            network=cls.network, name=cls.rand_name)
        cls.subnet_v6 = cls.create_subnet(
            network=cls.network, name=cls.rand_name, ip_version=6)
        cls.router = cls.create_router_by_client()
        cls.create_router_interface(cls.router['id'], cls.subnet_v4['id'])
        cls.create_router_interface(cls.router['id'], cls.subnet_v6['id'])
        cls.keypair = cls.create_keypair(name=cls.rand_name)
        cls.security_group = cls.create_security_group(name=cls.rand_name)
        cls.create_loginable_secgroup_rule(cls.security_group['id'])

    def _create_server_with_network(self, network, use_advanced_image=False):
        port = self._create_server_port(network=network)
        floating_ip = self.create_floatingip(port=port)
        ssh_client = self._create_ssh_client(
            floating_ip=floating_ip, use_advanced_image=use_advanced_image)
        server = self._create_server(port=port,
                                     use_advanced_image=use_advanced_image)
        return Server(
            floating_ip=floating_ip, server=server, ssh_client=ssh_client)

    def _create_server_port(self, network=None, **params):
        network = network or self.network
        return self.create_port(network=network, name=self.rand_name,
                                security_groups=[self.security_group['id']],
                                **params)

    def _create_server(self, port, use_advanced_image=False, **params):
        if use_advanced_image:
            flavor_ref = CONF.neutron_plugin_options.advanced_image_flavor_ref
            image_ref = CONF.neutron_plugin_options.advanced_image_ref
        else:
            flavor_ref = CONF.compute.flavor_ref
            image_ref = CONF.compute.image_ref
        return self.create_server(flavor_ref=flavor_ref,
                                  image_ref=image_ref,
                                  key_name=self.keypair['name'],
                                  networks=[{'port': port['id']}],
                                  **params)['server']

    def _create_ssh_client(self, floating_ip, use_advanced_image=False):
        if use_advanced_image:
            username = CONF.neutron_plugin_options.advanced_image_ssh_user
        else:
            username = CONF.validation.image_ssh_user
        return ssh.Client(host=floating_ip['floating_ip_address'],
                          username=username,
                          pkey=self.keypair['private_key'])

    def _assert_has_ssh_connectivity(self, ssh_client):
        ssh_client.exec_command('true')

    def _get_primary_interface(self, ssh_client):
        out = ssh_client.exec_command(
            "ip -6 -br address show scope link up | head -1 | cut -d ' ' -f1")
        interface = out.strip()
        if not interface:
            self.fail(
                'Could not find a single interface '
                'with an IPv6 link-local address.')
        return interface

    @testtools.skipUnless(
        CONF.neutron_plugin_options.advanced_image_ref or
        CONF.neutron_plugin_options.default_image_is_advanced,
        'Advanced image is required to run this test.')
    @decorators.idempotent_id('e680949a-f1cc-11ea-b49a-cba39bbbe5ad')
    def test_metadata_routed(self):
        use_advanced_image = (
            not CONF.neutron_plugin_options.default_image_is_advanced)

        vm = self._create_server_with_network(
            self.network, use_advanced_image=use_advanced_image)
        self.wait_for_server_active(server=vm.server)
        self.wait_for_guest_os_ready(vm.server)
        self.check_connectivity(host=vm.floating_ip['floating_ip_address'],
                                ssh_client=vm.ssh_client)
        interface = self._get_primary_interface(vm.ssh_client)

        try:
            out = vm.ssh_client.exec_command(
                'curl http://[%(address)s%%25%(interface)s]/' % {
                    'address': nlib_const.METADATA_V6_IP,
                    'interface': interface})
            self.assertIn('latest', out)

            out = vm.ssh_client.exec_command(
                'curl http://[%(address)s%%25%(interface)s]/openstack/' % {
                    'address': nlib_const.METADATA_V6_IP,
                    'interface': interface})
            self.assertIn('latest', out)
        except exceptions.SSHExecCommandFailed:
            self._log_console_output()
            self._log_local_network_status()



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from neutron_lib import constants
from tempest.lib import decorators
from tempest.lib import exceptions as lib_exc

from neutron_tempest_plugin.api import base
from neutron_tempest_plugin import config

CONF = config.CONF


class NetworksTestDHCPv6(base.BaseNetworkTest):
    _ip_version = 6

    def setUp(self):
        super(NetworksTestDHCPv6, self).setUp()
        self.addCleanup(self._clean_network)

    @classmethod
    def skip_checks(cls):
        super(NetworksTestDHCPv6, cls).skip_checks()
        msg = None
        if not CONF.network_feature_enabled.ipv6:
            msg = "IPv6 is not enabled"
        elif not CONF.network_feature_enabled.ipv6_subnet_attributes:
            msg = "DHCPv6 attributes are not enabled."
        if msg:
            raise cls.skipException(msg)

    @classmethod
    def resource_setup(cls):
        super(NetworksTestDHCPv6, cls).resource_setup()
        cls.network = cls.create_network()

    def _remove_from_list_by_index(self, things_list, elem):
        for index, i in enumerate(things_list):
            if i['id'] == elem['id']:
                del things_list[index]
                return

    def _clean_network(self):
        body = self.client.list_ports()
        ports = body['ports']
        for port in ports:
            if (port['device_owner'].startswith(
                    constants.DEVICE_OWNER_ROUTER_INTF) and
                port['device_id'] in [r['id'] for r in self.routers]):
                self.client.remove_router_interface_with_port_id(
                    port['device_id'], port['id']
                )
            else:
                if port['id'] in [p['id'] for p in self.ports]:
                    self.client.delete_port(port['id'])
                    self._remove_from_list_by_index(self.ports, port)
        body = self.client.list_subnets()
        subnets = body['subnets']
        for subnet in subnets:
            if subnet['id'] in [s['id'] for s in self.subnets]:
                self.client.delete_subnet(subnet['id'])
                self._remove_from_list_by_index(self.subnets, subnet)
        body = self.client.list_routers()
        routers = body['routers']
        for router in routers:
            if router['id'] in [r['id'] for r in self.routers]:
                self.client.delete_router(router['id'])
                self._remove_from_list_by_index(self.routers, router)

    @decorators.idempotent_id('98244d88-d990-4570-91d4-6b25d70d08af')
    def test_dhcp_stateful_fixedips_outrange(self):
        """Test DHCP Stateful fixed IPs out of range

        When port gets IP address from fixed IP range it
        shall be checked if it's from subnets range.
        """
        kwargs = {'ipv6_ra_mode': 'dhcpv6-stateful',
                  'ipv6_address_mode': 'dhcpv6-stateful'}
        subnet = self.create_subnet(self.network, **kwargs)
        ip_range = netaddr.IPRange(subnet["allocation_pools"][0]["start"],
                                   subnet["allocation_pools"][0]["end"])
        for i in range(1, 3):
            ip = netaddr.IPAddress(ip_range.last + i).format()
            self.assertRaises(lib_exc.BadRequest,
                              self.create_port,
                              self.network,
                              fixed_ips=[{'subnet_id': subnet['id'],
                                          'ip_address': ip}])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.common import utils as tutils
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators

from neutron_tempest_plugin.api import base_routers as base


class RoutersTestHA(base.BaseRouterTest):

    required_extensions = ['router', 'l3-ha']
    HA_NETWORK_NAME_TEMPL = "HA network tenant %s"

    @classmethod
    def resource_setup(cls):
        # The check above will pass if api_extensions=all, which does
        # not mean "l3-ha" extension itself is present.
        # Instead, we have to check whether "ha" is actually present by using
        # admin credentials to create router with ha=True attribute
        # and checking for BadRequest exception and that the resulting router
        # has a high availability attribute.
        super(RoutersTestHA, cls).resource_setup()
        name = data_utils.rand_name('pretest-check')
        router = cls.admin_client.create_router(name)
        cls.admin_client.delete_router(router['router']['id'])
        if 'ha' not in router['router']:
            msg = "'ha' attribute not found. HA Possibly not enabled"
            raise cls.skipException(msg)

    @decorators.idempotent_id('8abc177d-14f1-4018-9f01-589b299cbee1')
    def test_ha_router_creation(self):
        """Test HA router creation

        Test uses administrative credentials to create a
        HA (High Availability) router using the ha=True.

        Acceptance
        The router is created and the "ha" attribute is set to True
        """
        name = data_utils.rand_name('router')
        router = self._create_admin_router(name, ha=True)
        self.assertTrue(router['ha'])

    @decorators.idempotent_id('97b5f7ef-2192-4fa3-901e-979cd5c1097a')
    def test_legacy_router_creation(self):
        """Test legacy router creation

        Test uses administrative credentials to create a
        SF (Single Failure) router using the ha=False.

        Acceptance
        The router is created and the "ha" attribute is
        set to False, thus making it a "Single Failure Router"
        as opposed to a "High Availability Router"
        """
        name = data_utils.rand_name('router')
        router = self._create_admin_router(name, ha=False)
        self.assertFalse(router['ha'])

    @decorators.idempotent_id('5a6bfe82-5b23-45a4-b027-5160997d4753')
    def test_legacy_router_update_to_ha(self):
        """Test legacy to HA router update

        Test uses administrative credentials to create a
        SF (Single Failure) router using the ha=False.
        Then it will "update" the router ha attribute to True

        Acceptance
        The router is created and the "ha" attribute is
        set to False. Once the router is updated, the ha
        attribute will be set to True
        """
        name = data_utils.rand_name('router')
        # router needs to be in admin state down in order to be upgraded to HA
        router = self._create_admin_router(name, ha=False,
                                           admin_state_up=False)
        self.assertFalse(router['ha'])
        router = self.admin_client.update_router(router['id'],
                                                 ha=True)
        self.assertTrue(router['router']['ha'])

    @decorators.idempotent_id('0d8c0c8f-3809-4acc-a2c8-e0941333ff6c')
    @tutils.requires_ext(extension="provider", service="network")
    def test_delete_ha_router_keeps_ha_network_segment_data(self):
        """Test deleting an HA router keeps correct segment data for network.

        Each tenant with HA router has an HA network. The HA network is a
        normal tenant network with segmentation data like type (vxlan) and
        segmenation id. This test makes sure that after an HA router is
        deleted, those segmentation data are kept in HA network. This tests
        regression of https://bugs.launchpad.net/neutron/+bug/1732543.
        """
        for i in range(2):
            router = self._create_admin_router(
                data_utils.rand_name('router%d' % i),
                ha=True)
        ha_net_name = self.HA_NETWORK_NAME_TEMPL % router['tenant_id']
        ha_network_pre_delete = self.admin_client.list_networks(
            name=ha_net_name)['networks'][0]
        segmentation_id = ha_network_pre_delete['provider:segmentation_id']
        self._delete_router(router['id'], self.admin_client)

        ha_network_post_delete = self.admin_client.show_network(
            ha_network_pre_delete['id'])['network']
        self.assertEqual(
            ha_network_post_delete['provider:segmentation_id'],
            segmentation_id)



# Copyright (c) 2017 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from oslo_config import cfg
import testtools

from neutron_lib.utils import test
from tempest.common import utils
from tempest.common import waiters
from tempest.lib.common import ssh
from tempest.lib.common.utils import data_utils
from tempest.lib import decorators

from neutron_tempest_plugin import config
from neutron_tempest_plugin.scenario import constants
from neutron_tempest_plugin.vpnaas.scenario import base_vpnaas as base


CONF = config.CONF

# NOTE(huntxu): This is a workaround due to a upstream bug [1].
# VPNaaS 4in6 and 6in4 is not working properly with LibreSwan 3.19+.
# In OpenStack zuul checks the base CentOS 7 node is using Libreswan 3.20 on
# CentOS 7.4. So we need to provide a way to skip the 4in6 and 6in4 test cases
# for zuul.
#
# Once the upstream bug gets fixed and the base node uses a newer version of
# Libreswan with that fix, we can remove this.
#
# [1] https://github.com/libreswan/libreswan/issues/175
CONF.register_opt(
    cfg.BoolOpt('skip_4in6_6in4_tests',
                default=False,
                help='Whether to skip 4in6 and 6in4 test cases.'),
    'neutron_vpnaas_plugin_options'
)


class Vpnaas(base.BaseTempestTestCase):
    """Test the following topology

          +-------------------+
          | public            |
          | network           |
          |                   |
          +-+---------------+-+
            |               |
            |               |
    +-------+-+           +-+-------+
    | LEFT    |           | RIGHT   |
    | router  | <--VPN--> | router  |
    |         |           |         |
    +----+----+           +----+----+
         |                     |
    +----+----+           +----+----+
    | LEFT    |           | RIGHT   |
    | network |           | network |
    |         |           |         |
    +---------+           +---------+
    """

    credentials = ['primary', 'admin']
    inner_ipv6 = False
    outer_ipv6 = False

    @classmethod
    @utils.requires_ext(extension="vpnaas", service="network")
    def resource_setup(cls):
        super(Vpnaas, cls).resource_setup()

        # common
        cls.keypair = cls.create_keypair()
        cls.secgroup = cls.os_primary.network_client.create_security_group(
            name=data_utils.rand_name('secgroup-'))['security_group']
        cls.security_groups.append(cls.secgroup)
        cls.create_loginable_secgroup_rule(secgroup_id=cls.secgroup['id'])
        cls.create_pingable_secgroup_rule(secgroup_id=cls.secgroup['id'])
        cls.ikepolicy = cls.create_ikepolicy(
            data_utils.rand_name("ike-policy-"))
        cls.ipsecpolicy = cls.create_ipsecpolicy(
            data_utils.rand_name("ipsec-policy-"))

        cls.extra_subnet_attributes = {}
        if cls.inner_ipv6:
            cls.create_v6_pingable_secgroup_rule(
                secgroup_id=cls.secgroup['id'])
            cls.extra_subnet_attributes['ipv6_address_mode'] = 'slaac'
            cls.extra_subnet_attributes['ipv6_ra_mode'] = 'slaac'

        # LEFT
        cls.router = cls.create_router(
            data_utils.rand_name('left-router'),
            admin_state_up=True,
            external_network_id=CONF.network.public_network_id)
        cls.network = cls.create_network(network_name='left-network')
        ip_version = 6 if cls.inner_ipv6 else 4
        v4_cidr = netaddr.IPNetwork('10.20.0.0/24')
        v6_cidr = netaddr.IPNetwork('2001:db8:0:2::/64')
        cidr = v6_cidr if cls.inner_ipv6 else v4_cidr
        cls.subnet = cls.create_subnet(
            cls.network, ip_version=ip_version, cidr=cidr, name='left-subnet',
            **cls.extra_subnet_attributes)
        cls.create_router_interface(cls.router['id'], cls.subnet['id'])

        # Gives an internal IPv4 subnet for floating IP to the left server,
        # we use it to ssh into the left server.
        if cls.inner_ipv6:
            v4_subnet = cls.create_subnet(
                cls.network, ip_version=4, name='left-v4-subnet')
            cls.create_router_interface(cls.router['id'], v4_subnet['id'])

        # RIGHT
        cls._right_network, cls._right_subnet, cls._right_router = \
            cls._create_right_network()

    @classmethod
    def create_v6_pingable_secgroup_rule(cls, secgroup_id=None, client=None):
        # NOTE(huntxu): This method should be moved into the base class, along
        # with the v4 version.
        """This rule is intended to permit inbound ping6"""

        rule_list = [{'protocol': 'ipv6-icmp',
                      'direction': 'ingress',
                      'port_range_min': 128,  # type
                      'port_range_max': 0,  # code
                      'ethertype': 'IPv6',
                      'remote_ip_prefix': '::/0'}]
        client = client or cls.os_primary.network_client
        cls.create_secgroup_rules(rule_list, client=client,
                                  secgroup_id=secgroup_id)

    @classmethod
    def _create_right_network(cls):
        router = cls.create_router(
            data_utils.rand_name('right-router'),
            admin_state_up=True,
            external_network_id=CONF.network.public_network_id)
        network = cls.create_network(network_name='right-network')
        v4_cidr = netaddr.IPNetwork('10.10.0.0/24')
        v6_cidr = netaddr.IPNetwork('2001:db8:0:1::/64')
        cidr = v6_cidr if cls.inner_ipv6 else v4_cidr
        ip_version = 6 if cls.inner_ipv6 else 4
        subnet = cls.create_subnet(
            network, ip_version=ip_version, cidr=cidr, name='right-subnet',
            **cls.extra_subnet_attributes)
        cls.create_router_interface(router['id'], subnet['id'])

        return network, subnet, router

    def _create_server(self, create_floating_ip=True, network=None):
        if network is None:
            network = self.network
        port = self.create_port(network, security_groups=[self.secgroup['id']])
        if create_floating_ip:
            fip = self.create_and_associate_floatingip(port['id'])
        else:
            fip = None
        server = self.create_server(
            flavor_ref=CONF.compute.flavor_ref,
            image_ref=CONF.compute.image_ref,
            key_name=self.keypair['name'],
            networks=[{'port': port['id']}])['server']
        waiters.wait_for_server_status(self.os_primary.servers_client,
                                       server['id'],
                                       constants.SERVER_STATUS_ACTIVE)
        return {'port': port, 'fip': fip, 'server': server}

    def _setup_vpn(self):
        sites = [
            dict(name="left", network=self.network, subnet=self.subnet,
                 router=self.router),
            dict(name="right", network=self._right_network,
                 subnet=self._right_subnet, router=self._right_router),
        ]
        psk = data_utils.rand_name('mysecret')
        for i in range(0, 2):
            site = sites[i]
            site['vpnservice'] = self.create_vpnservice(
                site['subnet']['id'], site['router']['id'],
                name=data_utils.rand_name('%s-vpnservice' % site['name']))
        for i in range(0, 2):
            site = sites[i]
            vpnservice = site['vpnservice']
            peer = sites[1 - i]
            if self.outer_ipv6:
                peer_address = peer['vpnservice']['external_v6_ip']
                if not peer_address:
                    msg = "Public network must have an IPv6 subnet."
                    raise self.skipException(msg)
            else:
                peer_address = peer['vpnservice']['external_v4_ip']
            self.create_ipsec_site_connection(
                self.ikepolicy['id'],
                self.ipsecpolicy['id'],
                vpnservice['id'],
                peer_address=peer_address,
                peer_id=peer_address,
                peer_cidrs=[peer['subnet']['cidr']],
                psk=psk,
                name=data_utils.rand_name(
                    '%s-ipsec-site-connection' % site['name']))

    def _get_ip_on_subnet_for_port(self, port, subnet_id):
        for fixed_ip in port['fixed_ips']:
            if fixed_ip['subnet_id'] == subnet_id:
                return fixed_ip['ip_address']
        msg = "Cannot get IP address on specified subnet %s for port %r." % (
            subnet_id, port)
        raise self.fail(msg)

    @test.unstable_test("bug 1882220")
    def _test_vpnaas(self):
        # RIGHT
        right_server = self._create_server(network=self._right_network,
            create_floating_ip=False)
        right_ip = self._get_ip_on_subnet_for_port(
            right_server['port'], self._right_subnet['id'])

        # LEFT
        left_server = self._create_server()
        ssh_client = ssh.Client(left_server['fip']['floating_ip_address'],
                                CONF.validation.image_ssh_user,
                                pkey=self.keypair['private_key'],
                                ssh_key_type=CONF.validation.ssh_key_type)

        # check LEFT -> RIGHT connectivity via VPN
        self.check_remote_connectivity(ssh_client, right_ip,
                                       should_succeed=False)
        self._setup_vpn()
        self.check_remote_connectivity(ssh_client, right_ip)

        # Test VPN traffic and floating IP traffic don't interfere each other.
        if not self.inner_ipv6:
            # Assign a floating-ip and check connectivity.
            # This is NOT via VPN.
            fip = self.create_and_associate_floatingip(
                right_server['port']['id'])
            self.check_remote_connectivity(ssh_client,
                                           fip['floating_ip_address'])

            # check LEFT -> RIGHT connectivity via VPN again, to ensure
            # the above floating-ip doesn't interfere the traffic.
            self.check_remote_connectivity(ssh_client, right_ip)


class Vpnaas4in4(Vpnaas):

    @decorators.idempotent_id('aa932ab2-63aa-49cf-a2a0-8ae71ac2bc24')
    def test_vpnaas(self):
        self._test_vpnaas()


class Vpnaas4in6(Vpnaas):
    outer_ipv6 = True

    @decorators.idempotent_id('2d5f18dc-6186-4deb-842b-051325bd0466')
    @testtools.skipUnless(CONF.network_feature_enabled.ipv6,
                          'IPv6 tests are disabled.')
    @testtools.skipIf(
        CONF.neutron_vpnaas_plugin_options.skip_4in6_6in4_tests,
        'VPNaaS 4in6 test is skipped.')
    def test_vpnaas_4in6(self):
        self._test_vpnaas()


class Vpnaas6in4(Vpnaas):
    inner_ipv6 = True

    @decorators.idempotent_id('10febf33-c5b7-48af-aa13-94b4fb585a55')
    @testtools.skipUnless(CONF.network_feature_enabled.ipv6,
                          'IPv6 tests are disabled.')
    @testtools.skipIf(
        CONF.neutron_vpnaas_plugin_options.skip_4in6_6in4_tests,
        'VPNaaS 6in4 test is skipped.')
    def test_vpnaas_6in4(self):
        self._test_vpnaas()


class Vpnaas6in6(Vpnaas):
    inner_ipv6 = True
    outer_ipv6 = True

    @decorators.idempotent_id('8b503ffc-aeb0-4938-8dba-73c7323e276d')
    @testtools.skipUnless(CONF.network_feature_enabled.ipv6,
                          'IPv6 tests are disabled.')
    def test_vpnaas_6in6(self):
        self._test_vpnaas()
